# -*- coding: utf-8 -*-

"""Console script for mlbench_cli."""
import configparser
import json
import os
import subprocess
import sys
from time import sleep
import urllib3
from urllib import request

import click
import yaml
from appdirs import user_data_dir
from kubernetes import client
from mlbench_core.api import ApiClient, MLBENCH_IMAGES, MLBENCH_BACKENDS
from pyhelm.chartbuilder import ChartBuilder
from pyhelm.tiller import Tiller
from tabulate import tabulate
import pickle
from pathlib import Path

GCLOUD_NVIDIA_DAEMONSET = (
    "https://raw.githubusercontent.com/"
    "GoogleCloudPlatform/container-engine-accelerators/"
    "stable/nvidia-driver-installer/cos/"
    "daemonset-preloaded.yaml"
)


urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
TILLER_MANIFEST_DEPLOYMENT = """apiVersion: extensions/v1beta1
kind: Deployment
metadata:
  creationTimestamp: null
  labels:
    app: helm
    name: tiller
  name: tiller-deploy
  namespace: kube-system
spec:
  replicas: 1
  strategy: {}
  template:
    metadata:
      creationTimestamp: null
      labels:
        app: helm
        name: tiller
    spec:
      serviceAccount: tiller
      containers:
      - env:
        - name: TILLER_NAMESPACE
          value: kube-system
        - name: TILLER_HISTORY_MAX
          value: '0'
        image: gcr.io/kubernetes-helm/tiller:v2.14.3
        imagePullPolicy: IfNotPresent
        livenessProbe:
          httpGet:
            path: /liveness
            port: 44135
          initialDelaySeconds: 1
          timeoutSeconds: 1
        name: tiller
        ports:
        - containerPort: 44134
          name: tiller
        - containerPort: 44135
          name: http
        readinessProbe:
          httpGet:
            path: /readiness
            port: 44135
          initialDelaySeconds: 1
          timeoutSeconds: 1
        resources: {}
status: {}"""

TILLER_MANIFEST_SERVICE = """apiVersion: v1
kind: Service
metadata:
  creationTimestamp: null
  labels:
    app: helm
    name: tiller
  name: tiller-deploy
  namespace: kube-system
spec:
  ports:
  - name: tiller
    port: 44134
    targetPort: tiller
  selector:
    app: helm
    name: tiller
  type: ClusterIP
status:
  loadBalancer: {}"""


@click.group()
def cli(args=None):
    """Console script for mlbench_cli."""
    return 0


@cli.command()
@click.argument("name", type=str)
@click.argument("num_workers", nargs=-1, type=int, metavar="num-workers")
@click.option("--gpu", "-g", default=False, type=bool, is_flag=True)
@click.option("--light", "-l", default=False, type=bool, is_flag=True)
@click.option("--dashboard-url", "-u", default=None, type=str)
def run(name, num_workers, gpu, light, dashboard_url):
    """Start a new run for a benchmark image"""
    current_run_inputs = {}

    last_run_inputs_dir_location = os.path.join(
        os.environ["HOME"], ".local", "share", "mlbench"
    )
    Path(last_run_inputs_dir_location).mkdir(parents=True, exist_ok=True)

    last_run_inputs_file_location = os.path.join(
        last_run_inputs_dir_location, "last_run_inputs.pkl"
    )

    try:
        last_run_inputs = pickle.load(open(last_run_inputs_file_location, "rb"))
    except FileNotFoundError as e:
        last_run_inputs = {}

    images = list(MLBENCH_IMAGES.keys())

    text_prompt = "Benchmark: \n\n"

    text_prompt += "\n".join("[{}]\t{}".format(i, t) for i, t in enumerate(images))
    text_prompt += "\n[{}]\tCustom Image".format(len(images))

    text_prompt += "\n\nSelection"

    selection = click.prompt(
        text_prompt,
        type=click.IntRange(0, len(images)),
        default=last_run_inputs.get("benchmark", 0),
    )
    current_run_inputs["benchmark"] = selection

    if selection == len(images):
        # run custom image
        image = click.prompt(
            "Image", type=str, default=last_run_inputs.get("image", None)
        )
        current_run_inputs["image"] = image
        image_command = click.prompt(
            "Command", type=str, default=last_run_inputs.get("image_command", None)
        )
        current_run_inputs["image_command"] = image_command
        benchmark = {
            "custom_image_name": image,
            "custom_image_command": image_command,
        }
    else:
        benchmark = {"image": images[selection]}

    # Backend Prompt
    text_prompt = "Backend: \n\n"
    text_prompt += "\n".join(
        "[{}]\t{}".format(i, t) for i, t in enumerate(MLBENCH_BACKENDS)
    )
    text_prompt += "\n[{}]\tCustom Backend".format(len(MLBENCH_BACKENDS))
    text_prompt += "\n\nSelection"

    selection = click.prompt(
        text_prompt,
        type=click.IntRange(0, len(MLBENCH_BACKENDS)),
        default=last_run_inputs.get("backend", 0),
    )
    current_run_inputs["backend"] = selection

    if selection == len(MLBENCH_BACKENDS):
        backend = click.prompt(
            "Backend", type=str, default=last_run_inputs.get("custom_backend", None)
        )
        current_run_inputs["custom_backend"] = backend
        run_on_all = click.confirm(
            "Run command on all nodes (otherwise just first node)",
            default=last_run_inputs.get("run_on_all", None),
        )
        current_run_inputs["run_on_all"] = run_on_all
        benchmark["custom_backend"] = backend
        benchmark["run_all_nodes"] = run_on_all
    else:
        benchmark["backend"] = MLBENCH_BACKENDS[selection]

    pickle.dump(current_run_inputs, open(last_run_inputs_file_location, "wb"))

    benchmark["gpu_enabled"] = gpu
    benchmark["light_target"] = light

    loaded = setup_client_from_config()

    client = ApiClient(in_cluster=False, url=dashboard_url, load_config=not loaded)

    results = []

    for num_w in num_workers:
        current_name = "{}-{}".format(name, num_w)

        res = client.create_run(current_name, num_w, **benchmark)
        results.append(res)

    for res in results:
        act_result = res.result()
        if act_result.status_code > 201:
            try:
                click.echo(
                    "Couldn't start run: {}".format(act_result.json()["message"])
                )
            except json.JSONDecodeError:
                print(str(act_result.text))
                click.echo(
                    "Couldn't start run: Status {} for request".format(
                        act_result.status_code
                    )
                )
            return

        click.echo("Run started with name {}".format(act_result.json()["name"]))


@cli.command()
@click.argument("name", type=str)
@click.option("--dashboard-url", "--u", default=None, type=str)
def status(name, dashboard_url):
    """Get the status of a benchmark run"""
    loaded = setup_client_from_config()

    client = ApiClient(in_cluster=False, url=dashboard_url, load_config=not loaded)

    ret = client.get_runs()
    runs = ret.result().json()

    try:
        run = next(r for r in runs if r["name"] == name)
    except StopIteration:
        click.echo("Run not found")
        return

    del run["job_id"]
    del run["job_metadata"]

    click.echo(tabulate([run], headers="keys"))

    loss = client.get_run_metrics(
        run["id"], metric_filter="val_global_loss @ 0", last_n=1
    )
    prec = client.get_run_metrics(
        run["id"], metric_filter="val_global_Prec@1 @ 0", last_n=1
    )

    loss = loss.result()
    prec = prec.result()

    if loss.status_code < 300 and "val_global_loss @ 0" in loss.json():
        val = loss.json()["val_global_loss @ 0"][0]
        click.echo(
            "Current Global Loss: {0:.2f} ({1})".format(
                float(val["value"]), val["date"]
            )
        )
    else:
        click.echo("No Validation Loss Data yet")
    if prec.status_code < 300 and "val_global_Prec@1 @ 0" in prec.json():
        val = prec.json()["val_global_Prec@1 @ 0"][0]
        click.echo(
            "Current Global Precision: {0:.2f} ({1})".format(
                float(val["value"]), val["date"]
            )
        )
    else:
        click.echo("No Validation Precision Data yet")


@cli.command()
def get_dashboard_url():
    """Returns the dashboard URL of the current cluster"""
    loaded = setup_client_from_config()

    if not loaded:
        click.echo("No Cluster config found")
        return

    client = ApiClient(in_cluster=False, load_config=False)

    click.echo(client.endpoint.replace("api/", ""))


@cli.command()
@click.argument("name", type=str)
@click.option("--dashboard-url", "--u", default=None, type=str)
def delete(name, dashboard_url):
    """Delete a benchmark run"""
    loaded = setup_client_from_config()

    client = ApiClient(in_cluster=False, url=dashboard_url, load_config=not loaded)

    ret = client.get_runs()
    runs = ret.result().json()

    try:
        run = next(r for r in runs if r["name"] == name)
    except StopIteration:
        click.echo("Run not found")
        return

    del run["job_id"]
    del run["job_metadata"]

    client.delete_run(run["id"])


@cli.command()
@click.argument("name", type=str)
@click.option("--output", "-o", default="results.zip", type=str)
@click.option("--dashboard-url", "-u", default=None, type=str)
def download(name, output, dashboard_url):
    """Download the results of a benchmark run"""
    loaded = setup_client_from_config()

    client = ApiClient(in_cluster=False, url=dashboard_url, load_config=not loaded)

    ret = client.get_runs()
    runs = ret.result().json()

    run = next(r for r in runs if r["name"] == name)

    ret = client.download_run_metrics(run["id"])

    with open(output, "wb") as f:
        f.write(ret.result().content)


@cli.group("delete-cluster")
def delete_cluster():
    pass


@delete_cluster.command("gcloud")
@click.argument("name", type=str)
@click.option("--project", "-p", default=None, type=str)
@click.option("--zone", "-z", default="europe-west1-b", type=str)
def delete_gcloud(name, zone, project):
    from google.cloud import container_v1
    import google.auth
    from google.auth.exceptions import DefaultCredentialsError

    try:
        credentials, default_project = google.auth.default()
    except DefaultCredentialsError:
        raise click.UsageError(
            "Couldn't find gcloud credentials. Install the gcloud"
            " sdk ( https://cloud.google.com/sdk/docs/quickstart-linux ) and "
            "run 'gcloud auth application-default login' to login and create "
            "your credentials."
        )

    if not project:
        project = default_project

    # create cluster
    gclient = container_v1.ClusterManagerClient()

    name_path = "projects/{}/locations/{}/".format(project, zone)

    cluster_path = "{}clusters/{}".format(name_path, name)

    response = gclient.delete_cluster(None, None, None, name=cluster_path)

    # wait for cluster to load
    while response.status < response.DONE:
        response = gclient.get_operation(
            None, None, None, name=name_path + "/" + response.name
        )
        sleep(1)

    if response.status != response.DONE:
        raise ValueError("Cluster deletion failed!")

    click.echo("Cluster deleted.")


@cli.group("create-cluster")
def create_cluster():
    pass


@create_cluster.command("gcloud")
@click.argument("num_workers", type=int, metavar="num-workers")
@click.argument("release", type=str)
@click.option("--kubernetes-version", "-k", type=str, default="1.13")
@click.option("--machine-type", "-t", default="n1-standard-4", type=str)
@click.option("--disk-size", "-d", default=50, type=int)
@click.option("--num-cpus", "-c", default=1, type=int)
@click.option("--num-gpus", "-g", default=0, type=int)
@click.option("--gpu-type", default="nvidia-tesla-p100", type=str)
@click.option("--zone", "-z", default="europe-west1-b", type=str)
@click.option("--project", "-p", default=None, type=str)
@click.option("--preemptible", "-e", is_flag=True)
@click.option("--custom-value", "-v", multiple=True)
def create_gcloud(
    num_workers,
    release,
    kubernetes_version,
    machine_type,
    disk_size,
    num_cpus,
    num_gpus,
    gpu_type,
    zone,
    project,
    preemptible,
    custom_value,
):
    from google.cloud import container_v1
    import google.auth
    from google.auth.exceptions import DefaultCredentialsError
    from googleapiclient import discovery, http

    try:
        credentials, default_project = google.auth.default()
    except DefaultCredentialsError:
        raise click.UsageError(
            "Couldn't find gcloud credentials. Install the gcloud"
            " sdk ( https://cloud.google.com/sdk/docs/quickstart-linux ) and "
            "run 'gcloud auth application-default login' to login and create "
            "your credentials."
        )

    assert num_workers >= 2, "Number of workers should be at least 2"

    if not project:
        project = default_project

    # create cluster
    gclient = container_v1.ClusterManagerClient()

    name = "{}-{}".format(release, num_workers)
    name_path = "projects/{}/locations/{}/".format(project, zone)

    extraargs = {}

    if num_gpus > 0:
        extraargs["accelerators"] = [
            container_v1.types.AcceleratorConfig(
                accelerator_count=num_gpus, accelerator_type=gpu_type
            )
        ]

    # delete existing firewall, if any
    firewalls = discovery.build("compute", "v1", cache_discovery=False).firewalls()

    existing_firewalls = firewalls.list(project=project).execute()
    fw_name = "{}-firewall".format(name)

    if any(f["name"] == fw_name for f in existing_firewalls["items"]):
        response = {}
        while not hasattr(response, "status"):
            try:
                response = firewalls.delete(project=project, firewall=fw_name).execute()
            except http.HttpError as e:
                if e.resp.status == 404:
                    response = {}
                    break
                click.echo("Wait for firewall to be available for deletion")
                sleep(5)
                response = {}
        while hasattr(response, "status") and response.status < response.DONE:
            response = gclient.get_operation(None, None, None, name=response.selfLink)
            sleep(1)

    # create cluster
    cluster = container_v1.types.Cluster(
        name=name,
        initial_node_count=num_workers,
        node_config=container_v1.types.NodeConfig(
            machine_type=machine_type,
            disk_size_gb=disk_size,
            preemptible=preemptible,
            oauth_scopes=["https://www.googleapis.com/auth/devstorage.full_control",],
            **extraargs,
        ),
        addons_config=container_v1.types.AddonsConfig(
            http_load_balancing=container_v1.types.HttpLoadBalancing(disabled=True,),
            horizontal_pod_autoscaling=container_v1.types.HorizontalPodAutoscaling(
                disabled=True,
            ),
            kubernetes_dashboard=container_v1.types.KubernetesDashboard(disabled=True,),
            network_policy_config=container_v1.types.NetworkPolicyConfig(
                disabled=False,
            ),
        ),
        logging_service=None,
        monitoring_service=None,
    )
    response = gclient.create_cluster(None, None, cluster, parent=name_path)

    # wait for cluster to load
    while response.status < response.DONE:
        response = gclient.get_operation(
            None, None, None, name=name_path + "/" + response.name
        )
        sleep(1)

    if response.status != response.DONE:
        raise ValueError("Cluster creation failed!")

    cluster = gclient.get_cluster(None, None, None, name=name_path + "/" + name)

    auth_req = google.auth.transport.requests.Request()
    credentials.refresh(auth_req)
    configuration = client.Configuration()
    configuration.host = f"https://{cluster.endpoint}:443"
    configuration.verify_ssl = False
    configuration.api_key = {"authorization": "Bearer " + credentials.token}
    client.Configuration.set_default(configuration)

    if num_gpus > 0:
        with request.urlopen(GCLOUD_NVIDIA_DAEMONSET) as r:
            dep = yaml.safe_load(r)
            dep["spec"]["selector"] = {
                "matchLabels": dep["spec"]["template"]["metadata"]["labels"]
            }
            dep = client.ApiClient()._ApiClient__deserialize(dep, "V1DaemonSet")
            k8s_client = client.AppsV1Api()
            k8s_client.create_namespaced_daemon_set("kube-system", body=dep)

    # create tiller service account
    client.CoreV1Api().create_namespaced_service_account(
        "kube-system",
        {
            "apiVersion": "v1",
            "kind": "ServiceAccount",
            "metadata": {
                "name": "tiller",
                "generateName": "tiller",
                "namespace": "kube-system",
            },
        },
    )

    client.RbacAuthorizationV1beta1Api().create_cluster_role_binding(
        {
            "apiVersion": "rbac.authorization.k8s.io/v1beta1",
            "kind": "ClusterRoleBinding",
            "metadata": {"name": "tiller"},
            "roleRef": {
                "apiGroup": "rbac.authorization.k8s.io",
                "kind": "ClusterRole",
                "name": "cluster-admin",
            },
            "subjects": [
                {"kind": "ServiceAccount", "name": "tiller", "namespace": "kube-system"}
            ],
        }
    )

    # deploy tiller
    tiller_service = yaml.safe_load(TILLER_MANIFEST_SERVICE)
    tiller_dep = yaml.safe_load(TILLER_MANIFEST_DEPLOYMENT)
    client.CoreV1Api().create_namespaced_service("kube-system", tiller_service)
    client.ExtensionsV1beta1Api().create_namespaced_deployment(
        "kube-system", tiller_dep
    )

    sleep(1)

    pods = client.CoreV1Api().list_namespaced_pod(
        namespace="kube-system", label_selector="app=helm"
    )

    tiller_pod = pods.items[0]

    while True:
        # Wait for tiller
        resp = client.CoreV1Api().read_namespaced_pod(
            namespace="kube-system", name=tiller_pod.metadata.name
        )
        if resp.status.phase != "Pending":
            break
        sleep(5)

    # kubernetes python doesn't currently support port forward
    # https://github.com/kubernetes-client/python/issues/166
    ports = 44134

    # resp = stream(
    #     client.CoreV1Api().connect_get_namespaced_pod_portforward,
    #     name=tiller_pod.metadata.name,
    #     namespace=tiller_pod.metadata.namespace,
    #     ports=ports
    #     )

    with subprocess.Popen(
        [
            "kubectl",
            "port-forward",
            "--namespace={}".format(tiller_pod.metadata.namespace),
            tiller_pod.metadata.name,
            "{0}:{0}".format(ports),
            "--server={}".format(configuration.host),
            "--token={}".format(credentials.token),
            "--insecure-skip-tls-verify=true",
        ]
    ) as portforward:

        sleep(5)
        # install chart
        tiller = Tiller("localhost")
        chart = ChartBuilder(
            {
                "name": "mlbench-helm",
                "source": {
                    "type": "git",
                    "location": "https://github.com/mlbench/mlbench-helm",
                },
            }
        )

        values = {
            "limits": {"workers": num_workers - 1, "gpu": num_gpus, "cpu": num_cpus}
        }

        if custom_value:
            # merge custom values with values
            for cv in custom_value:
                key, v = cv.split("=", 1)

                current = values
                key_path = key.split(".")

                for k in key_path[:-1]:
                    if k not in current:
                        current[k] = {}

                    current = current[k]

                current[key_path[-1]] = v

        tiller.install_release(
            chart.get_helm_chart(),
            name=name,
            wait=True,
            dry_run=False,
            namespace="default",
            values=values,
        )

        portforward.terminate()

    # open port in firewall
    mlbench_client = ApiClient(in_cluster=False, load_config=False)
    firewall_body = {
        "name": fw_name,
        "direction": "INGRESS",
        "sourceRanges": "0.0.0.0/0",
        "allowed": [{"IPProtocol": "tcp", "ports": [mlbench_client.port]}],
    }

    firewalls.insert(project=project, body=firewall_body).execute()

    config = get_config()

    config.set("general", "provider", "gke")

    config.set("gke", "cluster", cluster.endpoint)

    write_config(config)

    click.echo("MLBench successfully deployed")


def get_config_path():
    user_dir = user_data_dir("mlbench", "mlbench")
    return os.path.join(user_dir, "mlbench.ini")


def get_config():
    path = get_config_path()

    config = configparser.ConfigParser()

    if os.path.exists(path):
        config.read(path)

    if not config.has_section("general"):
        config.add_section("general")

    if not config.has_section("gke"):
        config.add_section("gke")

    return config


def write_config(config):
    path = get_config_path()

    if not os.path.exists(path):
        os.makedirs(os.path.dirname(path))

    with open(path, "w") as configfile:
        config.write(configfile)


def setup_client_from_config():
    config = get_config()

    provider = config.get("general", "provider", fallback=None)

    if not provider:
        return False

    if provider == "gke":
        return setup_gke_client_from_config(config)
    else:
        raise NotImplementedError()


def setup_gke_client_from_config(config):
    import google.auth
    from google.auth.exceptions import DefaultCredentialsError

    cluster = config.get("gke", "cluster")
    if not cluster:
        return False

    try:
        credentials, default_project = google.auth.default()
    except DefaultCredentialsError:
        raise click.UsageError(
            "Couldn't find gcloud credentials. Install the gcloud"
            " sdk ( https://cloud.google.com/sdk/docs/quickstart-linux ) and "
            "run 'gcloud auth application-default login' to login and create "
            "your credentials."
        )
    auth_req = google.auth.transport.requests.Request()
    credentials.refresh(auth_req)
    configuration = client.Configuration()
    configuration.host = f"https://{cluster}:443"
    configuration.verify_ssl = False
    configuration.api_key = {"authorization": "Bearer " + credentials.token}
    client.Configuration.set_default(configuration)

    return True


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



"""Utilities for measuring the performance of a model."""

from abc import abstractmethod

import torch
from mlbench_core.utils import AverageMeter
from mlbench_core.utils.pytorch.distributed import global_average

try:
    import sacrebleu
except ImportError as e:
    pass


class MLBenchMetric(object):
    def __init__(self):
        self.average_meter = AverageMeter()

    @abstractmethod
    def __call__(self, loss, output, target):
        pass

    def reset(self):
        self.average_meter = AverageMeter()

    def update(self, perc, size):
        self.average_meter.update(perc, size)

    def average(self):
        return global_average(self.average_meter.sum, self.average_meter.count)


class TopKAccuracy(MLBenchMetric):
    r"""Top K accuracy of an output.

    Counts a prediction as correct if the target value is in the top ``k``
    predictions, false otherwise, and returns the number of correct
    instances relative to total instances (0.0 to 100.0).

    Args:
        topk (int, optional): The number of top predictions to consider.
            Default: ``1``

    """

    def __init__(self, topk=1):
        super(TopKAccuracy, self).__init__()
        self.topk = topk

    def __call__(self, loss, output, target):
        """Computes the precision@k for the specified values of k

        Args:
            loss (:obj:`torch.Tensor`): Not used for accuracy
            output (:obj:`torch.Tensor`): Predictions of a model
            target (:obj:`torch.Tensor`): Target labels

        Example:
                >>> m = nn.Softmax()
                >>> input = torch.randn(10, 50)
                >>> preds = m(input)
                >>> targets = torch.randint(0, 1, (10,50))
                >>> topk = TopKAccuracy(5)
                >>> precision = topk(preds, targets)

        Returns:
            float
        """
        batch_size = target.size(0)

        output = self._preprocess_output(output)

        _, pred = output.topk(self.topk, 1, True, True)
        pred = pred.t().float()
        correct = pred.eq(target.view(1, -1).expand_as(pred).float())
        correct_k = correct[: self.topk].view(-1).float().sum(0, keepdim=True)
        return correct_k.mul_(100.0 / batch_size)

    def _preprocess_output(self, output):
        dim = output.size(1)
        if dim == 1:
            output = torch.cat((1 - output, output), 1)  # Increase dimension
            dim = output.size(1)
        if self.topk >= dim:
            raise ValueError(
                "Cannot compute top {} accuracy with "
                "input dimension {}".format(self.topk, dim)
            )
        return output

    @property
    def name(self):
        """str: Name of this metric."""
        return "Prec@{}".format(self.topk)


class Perplexity(MLBenchMetric):
    """Language Model Perplexity score."""

    @property
    def name(self):
        """str: Name of this metric."""
        return "Perplexity"

    def __call__(self, loss, output, target):
        """Computes the perplexity

        Args:
            loss (:obj:`torch.Tensor`): The loss of a language model.
            output (:obj:`torch.Tensor`): Not Used
            target (:obj:`torch.Tensor`): Not Used

        Returns:
            float
        """
        return torch.exp(loss)


class DiceCoefficient(MLBenchMetric):
    def __call__(self, loss, output, target):
        """ Computes the Dice Coefficient of a Binary classification problem

        Args:
            loss (:obj:`torch.Tensor`): Not Used
            output (:obj:`torch.Tensor`): Output of model
            target (:obj:`torch.Tensor`): Target labels

        Returns:
            float: Dice Coefficient in [0,1]
        """
        eps = 0.0001
        output, target = output.float(), target.float()
        self.inter = torch.dot(output.view(-1), target.view(-1))
        self.union = torch.sum(output) + torch.sum(target) + eps

        t = (2 * self.inter.float() + eps) / self.union.float()
        return t

    @property
    def name(self):
        """str: Name of this metric"""
        return "Dice Coefficient"


class F1Score(MLBenchMetric):
    def __init__(self, threshold=0.5, eps=1e-9):
        """ F1-Score metric

        Args:
            threshold (float): Threshold for prediction probability
        """
        super(F1Score, self).__init__()
        self.threshold = threshold
        self.eps = eps

    def __call__(self, loss, output, target):
        """ Computes the F1-Score of a Binary classification problem

        Args:
            loss (:obj:`torch.Tensor`): Not Used
            output (:obj:`torch.Tensor`): Output of model
            target (:obj:`torch.Tensor`): Target labels

        Returns:
            float: F1-Score in [0,1]
        """

        y_pred = torch.ge(output.float(), self.threshold).float()
        y_true = target.float()

        true_positive = (y_pred * y_true).sum(dim=0)
        precision = true_positive.div(y_pred.sum(dim=0).add(self.eps))
        recall = true_positive.div(y_true.sum(dim=0).add(self.eps))

        return torch.mean(
            (precision * recall).div(precision + recall + self.eps).mul(2)
        )

    @property
    def name(self):
        return "F1-Score"


class BLEUScore(MLBenchMetric):
    def __init__(self):
        """ Bilingual Evaluation Understudy score"""
        super(BLEUScore, self).__init__()

    def __call__(self, loss, output, target):
        """ Computes the BLEU score of a translation task

        Args:
            loss (:obj:`torch.Tensor`): Not Used
            output (:obj:`torch.Tensor`): Translated output (not tokenized)
            target (:obj:`torch.Tensor`): Target labels

        Returns:
            float: BLEU score
        """
        return torch.tensor(
            [
                sacrebleu.corpus_bleu(
                    output, [target], tokenize="intl", lowercase=True
                ).score
            ]
        )

    @property
    def name(self):
        return "BLEU-Score"



import torch
from mlbench_core.evaluation.pytorch.metrics import F1Score, TopKAccuracy
import numpy as np


def test_f1_score():
    output = torch.tensor([1, 1, 1, 1, 1]).reshape(5, 1)
    target = torch.tensor([0, 0, 0, 0, 0]).reshape(5, 1)

    f1 = F1Score()
    score = f1(None, output, target)

    assert score.item() == 0

    output = torch.tensor([1, 1, 1, 0, 1]).reshape(5, 1)
    target = torch.tensor([1, 0, 1, 1, 0]).reshape(5, 1)

    precision = 2 / (2 + 2)
    recall = 2 / (2 + 1)

    score = f1(None, output, target)
    expected_score = 2 * (precision * recall) / (precision + recall)
    np.testing.assert_almost_equal(score.item(), expected_score)


def test_top1_accuracy():
    output_1 = torch.tensor([[0, 1], [0, 1], [1, 0], [0, 1], [1, 0]]).reshape(5, 2)
    output_2 = torch.tensor([1, 1, 0, 1, 0]).reshape(5, 1)
    target = torch.tensor([0, 1, 0, 0, 1]).reshape(5, 1)

    acc = TopKAccuracy()
    expected_score = (2 / 5) * 100

    actual_score_1 = acc(None, output_1, target)
    actual_score_2 = acc(None, output_2, target)

    assert actual_score_1 == expected_score
    assert actual_score_2 == expected_score



r"""A controlflow which train and evaluate a model."""
import logging

from mlbench_core.utils import AverageMeter, Tracker


def train_round(
    session,
    train_set_init_op,
    train_op,
    loss_op,
    metrics,
    batch_size,
    num_batches_per_epoch_for_train,
    tracker,
    lr_scheduler_level=None,
    lr_tensor=None,
):
    """ Performs num_batches_per_epoch_for_train batches of training (or full trainset if
    not specified)

    Args:
        session (obj): The tensorflow session
        train_set_init_op (obj): The trainset initialisation tf operation
        train_op (obj): The tensorflow training operation
        loss_op (obj): The tensorflow loss operation
        metrics (list): List of metrics to track
        batch_size (int): The batch size
        num_batches_per_epoch_for_train (int): Maximum number of batches tot rain for per epoch,
                                   default: `None` (all batches)
        tracker (`obj`:mlbench_core.utils.Tracker): Tracker object to use
        lr_scheduler_level (str): Learning Rate scheduler mode, one of `batch` or `epoch`
        lr_tensor (obj): The learningrate schedule tensorflow operation
    """
    logging.info("Initialize training dataset.")
    session.run(train_set_init_op)
    tracker.train()

    loss_meter = AverageMeter()
    metrics_meter = [AverageMeter() for _ in metrics]

    if lr_scheduler_level == "epoch" and lr_tensor is not None:
        lr = session.run(lr_tensor)
        logging.debug(
            "Epoch {} Learning Rate : {:10.3e}".format(tracker.current_epoch, lr)
        )

    for i_batch in range(num_batches_per_epoch_for_train):
        # for i_batch in range(1):
        tracker.batch_start()

        if lr_scheduler_level == "batch" and lr_tensor is not None:
            lr = session.run(lr_tensor)
            logging.debug(
                "Epoch {} Learning Rate : {:10.3e}".format(tracker.current_epoch, lr)
            )

        out = session.run(
            {
                "metrics": [m.metric_op for m in metrics],
                "loss": loss_op,
                "train_op": train_op,
            }
        )

        tracker.batch_end()

        # Update tracker
        loss_meter.update(out["loss"], n=batch_size)
        tracker.record_loss(loss_meter.avg, log_to_api=True)

        for metric, meter, o in zip(metrics, metrics_meter, out["metrics"]):
            meter.update(o, n=batch_size)
            tracker.record_metric(metric, meter.avg, log_to_api=True)

        # Print logging information.
        progress = i_batch / num_batches_per_epoch_for_train
        progress += tracker.current_epoch

        status = "Epoch {:5.2f} Batch {:4}: ".format(progress, i_batch)

        logging.info(status + str(tracker))

    # Record training loss and metrics.
    tracker.record_loss(loss_meter.avg, log_to_api=True)

    for metric, meter in zip(metrics, metrics_meter):
        tracker.record_metric(metric, meter.avg, log_to_api=True)

    logging.info("Finish training for one epoch.")


def validation_round(
    session,
    validation_set_init_op,
    loss_op,
    metrics,
    batch_size,
    num_batches_per_epoch_for_validation,
    tracker,
):
    """ Handles one full iteration of validation on the whole validation set.

    Args:
        session (obj): The tensorflow session
        validation_set_init_op (obj): The trainset initialisation tf operation
        loss_op (obj): The tensorflow loss operation
        metrics (list): List of metrics to track
        batch_size (int): The batch size
        num_batches_per_epoch_for_validation (int): Maximum number of batches to validate
            for per epoch, default: `None` (all batches)
        tracker (`obj`:mlbench_core.utils.Tracker): Tracker object to use
    """
    session.run(validation_set_init_op)
    tracker.validation()

    loss_meter = AverageMeter()
    metrics_meter = [AverageMeter() for _ in metrics]

    for i_batch in range(num_batches_per_epoch_for_validation):
        out = session.run({"metrics": [m.metric_op for m in metrics], "loss": loss_op})

        # Update tracker
        loss_meter.update(out["loss"], n=batch_size)
        for meter, o in zip(metrics_meter, out["metrics"]):
            meter.update(o, n=batch_size)

        logging.debug(
            "{}/{} Validation loss={:10.3e} | metrics: [{}]".format(
                tracker.current_epoch,
                i_batch,
                loss_meter.avg,
                ",".join([format(m.avg, "10.3e") for m in metrics_meter]),
            )
        )

    tracker.record_loss(loss_meter.avg, log_to_api=True)

    if tracker.rank == 0:
        tracker.record_stat("global_loss", loss_meter.avg, log_to_api=True)

    for i, metric, meter in zip(range(len(metrics)), metrics, metrics_meter):
        tracker.record_metric(metric, meter.avg, log_to_api=True)

        if tracker.rank == 0:
            tracker.record_stat(
                "global_{}".format(metric.name), meter.avg, log_to_api=True
            )


class TrainValidation(object):
    """A control flow to train and evaluate a model.

    Args:
        train_op (:obj:`tf.Operation`): An operation for training models.
        sess (:obj:`tf.Session`): A session which the control flow will communicate.
        loss (:obj:`tf.Tensor`): The loss tensor.
        metrics (list of :obj:`tf.Tensor`): A list of metrics tensors.
        max_train_steps (int): Number of steps for training (independent of lr)
        train_epochs (int): Number of steps for training (may related to lr).
        batch_size (int): Size of a batch.
        num_batches_per_epoch_for_train (int): Number of batches in one training epoch
        num_batches_per_epoch_for_validation (int): Number of batches in one validation epoch
        train_set_init_op (:obj:`tf.Operation`): Op for initializing training dataset.
        validation_set_init_op (:obj:`tf.Operation`): Op for initializing validation dataset.
        run_id (str): the id of the run in the dashboard
        rank (int): the rank of the current worker
        lr_scheduler_level (str): Learning rate is updated based on `epoch` or `batch`.

    """

    def __init__(
        self,
        train_op,
        sess,
        loss,
        metrics,
        max_train_steps,
        train_epochs,
        batch_size,
        num_batches_per_epoch_for_train,
        num_batches_per_epoch_for_validation,
        train_set_init_op,
        validation_set_init_op,
        run_id,
        rank,
        lr_scheduler_level="epoch",
        tracker=None,
    ):
        self.batch_size = batch_size
        self.num_batches_per_epoch_for_train = num_batches_per_epoch_for_train
        self.num_batches_per_epoch_for_validation = num_batches_per_epoch_for_validation
        self.sess = sess
        self.loss = loss
        self.metrics = metrics
        self.train_op = train_op
        self.lr_scheduler_level = lr_scheduler_level
        self.max_train_steps = max_train_steps
        self.train_epochs = train_epochs
        self.train_set_init_op = train_set_init_op
        self.validation_set_init_op = validation_set_init_op
        self.run_id = run_id
        self.rank = rank
        if tracker:
            self.tracker = tracker
        else:
            self.tracker = Tracker(metrics, run_id, rank)

    def train_one_epoch(self, lr_tensor_name=None):
        """Train a model for an epoch and use tracker to log stats.

        Args:
            lr_tensor (obj): The learningrate schedule tensorflow operation"""
        train_round(
            self.sess,
            self.train_set_init_op,
            self.train_op,
            self.loss,
            self.metrics,
            self.batch_size,
            self.num_batches_per_epoch_for_train,
            self.tracker,
            lr_tensor=lr_tensor_name,
            lr_scheduler_level=self.lr_scheduler_level,
        )

    def valid_one_epoch(self):
        """Validate a model for an epoch and use tracker to log stats."""
        validation_round(
            self.sess,
            self.validation_set_init_op,
            self.loss,
            self.metrics,
            self.batch_size,
            self.num_batches_per_epoch_for_validation,
            self.tracker,
        )

    def train_and_eval(self, initial_epoch=0, lr_tensor_name=None):
        """Train and evaluate one epoch.

        Args:
            initial_epoch (int, optional): Defaults to 0. Initial epoch of training.
            lr_tensor_name (:obj:`tf.Tensor`, optional): Defaults to None.
                A (scalar) float tensor representing name of learning rate
        """
        final_epoch = min(self.max_train_steps, self.train_epochs)
        for i_epoch in range(initial_epoch, final_epoch):
            logging.debug("=> Epoch {}".format(i_epoch))

            self.train_one_epoch()
            self.valid_one_epoch()
            self.tracker.epoch_end()

        return self.tracker



from .tracker import AverageMeter, Tracker
from .log_metrics import LogMetrics

try:
    import torch
    from . import pytorch
except ImportError:
    pass


try:
    import tensorflow
    from . import tensorflow
except ImportError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `mlbench_core.utils.pytorch.helpers` package."""

import datetime

from freezegun import freeze_time
from mlbench_core.evaluation.goals import task1_time_to_accuracy_light_goal
from mlbench_core.evaluation.pytorch.metrics import TopKAccuracy
from mlbench_core.utils import Tracker


def test_tracker():
    tracker = Tracker([TopKAccuracy(5)], 1, 0)

    assert tracker is not None


def test_tracker_goal(mocker):
    patched = mocker.patch("mlbench_core.utils.tracker.LogMetrics")

    metric = TopKAccuracy(1)
    tracker = Tracker([metric], 1, 0, task1_time_to_accuracy_light_goal())

    tracker.start()

    assert tracker.start_time is not None

    tracker.train()

    tracker.record_stat("global_Prec@1", 69, log_to_api=True)
    tracker.batch_end()

    assert not tracker.goal_reached

    tracker.record_stat("global_Prec@1", 70, log_to_api=True)
    tracker.batch_end()

    assert not tracker.goal_reached

    tracker.validation()

    tracker.record_stat("global_Prec@1", 69, log_to_api=True)
    tracker.batch_end()

    assert not tracker.goal_reached

    tracker.record_stat("global_Prec@1", 70, log_to_api=True)

    assert tracker.goal_reached


def test_tracker_goal_times(mocker):
    patched = mocker.patch("mlbench_core.utils.tracker.LogMetrics")

    metric = TopKAccuracy(1)
    tracker = Tracker([metric], 1, 0, task1_time_to_accuracy_light_goal())

    tracker.start()

    assert tracker.start_time is not None

    tracker.train()

    with freeze_time(datetime.datetime.now()) as frozen:
        tracker.batch_start()
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.record_batch_step("init")
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.record_batch_step("fwd_pass")
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.record_batch_step("comp_loss")
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.record_batch_step("backprop")
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.record_batch_step("opt_step")
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.batch_end()

        assert abs(tracker.get_total_communication_time() - 0.5) < 0.01
        assert abs(tracker.get_total_compute_time() - 1.5) < 0.01

        tracker.batch_start()
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.record_batch_step("init")
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.record_batch_step("fwd_pass")
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.record_batch_step("comp_loss")
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.record_batch_step("backprop")
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.record_batch_step("opt_step")
        frozen.tick(delta=datetime.timedelta(seconds=0.5))
        tracker.batch_end()

        assert abs(tracker.get_total_communication_time() - 1.0) < 0.01
        assert abs(tracker.get_total_compute_time() - 3.0) < 0.01

        tracker.validation()
        tracker.record_stat("global_Prec@1", 70, log_to_api=True)

        assert tracker.goal_reached
        assert any(filter(lambda c: c[1][3] == "TaskResult", patched.method_calls))



import torch.distributed as dist
from .helpers import config_logging
from .helpers import config_pytorch
from .helpers import config_path
from .helpers import Timeit
from .topology import FCGraph

from contextlib import contextmanager

import os
import torch

__all__ = ["initialize_backends", "Timeit", "FCGraph"]


@contextmanager
def initialize_backends(
    comm_backend="mpi",
    hosts=None,
    rank=-1,
    logging_level="INFO",
    logging_file="/mlbench.log",
    use_cuda=False,
    seed=None,
    cudnn_deterministic=False,
    ckpt_run_dir="/checkpoints",
    delete_existing_ckpts=False,
):
    """Initializes the backends.

    Sets up logging, sets up pytorch and configures paths
    correctly.

    Args:
        config (:obj:`types.SimpleNamespace`): a global object containing all of the config.

    Returns:
        (:obj:`types.SimpleNamespace`): a global object containing all of the config.
    """

    if not (hasattr(dist, "_initialized") and dist._initialized):

        if comm_backend in [dist.Backend.GLOO, dist.Backend.NCCL]:

            if comm_backend == dist.Backend.NCCL:
                pass
                # assert (
                #     torch.cuda.is_available()
                # ), "Invalid use of NCCL backend without CUDA support available"

            hosts = hosts.split(",")
            os.environ["MASTER_ADDR"] = hosts[0]
            os.environ["MASTER_PORT"] = "29500"
            os.environ["RANK"] = str(rank)
            os.environ["WORLD_SIZE"] = str(len(hosts))

        dist.init_process_group(comm_backend)

    config_logging(logging_level, logging_file)

    rank, world_size, graph = config_pytorch(use_cuda, seed, cudnn_deterministic)

    config_path(ckpt_run_dir, delete_existing_ckpts)

    yield rank, world_size, graph

    dist.destroy_process_group()



import logging
import os

import torchvision.datasets as datasets
import torchvision.transforms as transforms

_logger = logging.getLogger("mlbench")


class CIFAR10V1(datasets.CIFAR10):
    """CIFAR10V1 Dataset.

    Loads CIFAR10V1 images with mean and std-dev normalisation.
    Performs random crop and random horizontal flip on train and
    only normalisation on val.
    Based on `torchvision.datasets.CIFAR10` and `Pytorch CIFAR 10 Example`_.

    Args:
        root (str): Root folder for the dataset
        train (bool): Whether to get the train or validation set (default=True)
        download (bool): Whether to download the dataset if it's not present

    .. _Pytorch CIFAR 10 Example:
       https://github.com/kuangliu/pytorch-cifar/blob/master/main.py
    """

    def __init__(self, root, train=True, download=False):
        cifar10_stats = {
            "mean": (0.4914, 0.4822, 0.4465),
            "std": (0.2023, 0.1994, 0.2010),
        }

        if train:
            transform = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.RandomCrop(32, padding=4),
                    transforms.ToTensor(),
                    transforms.Normalize(cifar10_stats["mean"], cifar10_stats["std"]),
                ]
            )
        else:
            transform = transforms.Compose(
                [
                    transforms.ToTensor(),
                    transforms.Normalize(cifar10_stats["mean"], cifar10_stats["std"]),
                ]
            )
        super(CIFAR10V1, self).__init__(
            root=root, train=train, transform=transform, download=download
        )


class Imagenet(datasets.ImageFolder):
    """Imagenet (ILSVRC2017) Dataset.

    Loads Imagenet images with mean and std-dev normalisation.
    Performs random crop and random horizontal flip on train and
    resize + center crop on val.
    Based on `torchvision.datasets.ImageFolder`

    Args:
        root (str): Root folder of Imagenet dataset (without `train/` or `val/`)
        train (bool): Whether to get the train or validation set (default=True)
    """

    def __init__(self, root, train=True):
        self.train = train

        imagenet_stats = {"mean": [0.485, 0.456, 0.406], "std": [0.229, 0.224, 0.225]}

        if train:
            transform = transforms.Compose(
                [
                    transforms.RandomResizedCrop(224),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Normalize(imagenet_stats["mean"], imagenet_stats["std"]),
                ]
            )
            self.root = os.path.join(self.root, "train")
        else:
            transform = transforms.Compose(
                [
                    transforms.Resize(256),
                    transforms.CenterCrop(224),
                    transforms.ToTensor(),
                    transforms.Normalize(imagenet_stats["mean"], imagenet_stats["std"]),
                ]
            )
            self.root = os.path.join(self.root, "val")

        super().__init__(self.root, transform)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `mlbench_core.api` package."""

import pytest
import datetime

from mlbench_core.api import ApiClient


@pytest.fixture
def kubernetes_api_client_node_port(mocker):
    mock_client = mocker.patch("kubernetes.client.CoreV1Api")
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.spec.type = (
        "NodePort"
    )
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__len__.return_value = (
        1
    )
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.spec.ports.__iter__.return_value = [
        mocker.MagicMock(node_port=12345, port=80)
    ]
    mock_client.return_value.list_namespaced_pod.return_value.items.__len__.return_value = (
        1
    )
    mock_client.return_value.read_node.return_value.status.addresses.__len__.return_value = (
        1
    )
    mock_client.return_value.read_node.return_value.status.addresses.__iter__.return_value = [
        mocker.MagicMock(type="ExternalIP", address="1.1.1.1")
    ]

    return mock_client


@pytest.fixture
def kubernetes_api_client_node_port_internal(mocker):
    mock_client = mocker.patch("kubernetes.client.CoreV1Api")
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.spec.type = (
        "NodePort"
    )
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__len__.return_value = (
        1
    )
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.spec.ports.__iter__.return_value = [
        mocker.MagicMock(node_port=12345, port=80)
    ]
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.spec.ports.__len__.return_value = (
        1
    )
    mock_client.return_value.list_namespaced_pod.return_value.items.__len__.return_value = (
        1
    )
    mock_client.return_value.read_node.return_value.status.addresses.__len__.return_value = (
        1
    )
    mock_client.return_value.read_node.return_value.status.addresses.__iter__.return_value = [
        mocker.MagicMock(type="InternalIP", address="1.1.1.1")
    ]

    return mock_client


@pytest.fixture
def kubernetes_api_client_clusterip(mocker):
    mock_client = mocker.patch("kubernetes.client.CoreV1Api")
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.spec.type = (
        "ClusterIP"
    )
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__len__.return_value = (
        1
    )
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.spec.cluster_ip = (
        "1.1.1.1"
    )
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.spec.ports.__getitem__.return_value.port = (
        12345
    )

    return mock_client


@pytest.fixture
def kubernetes_api_client_loadbalancer(mocker):
    mock_client = mocker.patch("kubernetes.client.CoreV1Api")
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.spec.type = (
        "LoadBalancer"
    )
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__len__.return_value = (
        1
    )
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.spec.ports.__getitem__.return_value.port = (
        12345
    )
    mock_client.return_value.list_service_for_all_namespaces.return_value.items.__getitem__.return_value.status.load_balancer.ingress.ip = (
        "1.1.1.1"
    )

    return mock_client


@pytest.fixture
def kubernetes_api_client_incluster(mocker):
    mock_client = mocker.patch("kubernetes.client.CoreV1Api")
    mock_client.return_value.list_namespaced_pod.return_value.items.__len__.return_value = (
        1
    )
    mock_client.return_value.list_namespaced_pod.return_value.items.__getitem__.return_value.status.pod_ip = (
        "1.1.1.1"
    )

    return mock_client


def test_instantiation(mocker, kubernetes_api_client_node_port):
    mocker.patch("kubernetes.config.load_kube_config")
    with ApiClient(in_cluster=False) as client:
        assert client is not None
        assert client.endpoint == "http://1.1.1.1:12345/api/"


def test_instantiation_nodeport_internal(
    mocker, kubernetes_api_client_node_port_internal
):
    mocker.patch("kubernetes.config.load_kube_config")
    client = ApiClient(in_cluster=False)

    assert client is not None
    assert client.endpoint == "http://1.1.1.1:12345/api/"


def test_instantiation_url():
    client = ApiClient(url="1.1.1.1:12345")

    assert client is not None
    assert client.endpoint == "http://1.1.1.1:12345/api/"


def test_instantiation_incluster(mocker, kubernetes_api_client_incluster):
    mocker.patch("kubernetes.config.load_incluster_config")

    client = ApiClient(in_cluster=True)

    assert client is not None
    assert client.endpoint == "http://1.1.1.1:80/api/"


def test_instantiation_clusterip(mocker, kubernetes_api_client_clusterip):
    mocker.patch("kubernetes.config.load_kube_config")
    client = ApiClient(in_cluster=False)

    assert client is not None
    assert client.endpoint == "http://1.1.1.1:12345/api/"


def test_instantiation_loadbalancer(mocker, kubernetes_api_client_loadbalancer):
    mocker.patch("kubernetes.config.load_kube_config")
    client = ApiClient(in_cluster=False)

    assert client is not None
    assert client.endpoint == "http://1.1.1.1:12345/api/"


def test_get_all_metrics(mocker, kubernetes_api_client_node_port):
    mocker.patch("kubernetes.config.load_kube_config")
    rg = mocker.patch("concurrent.futures.ThreadPoolExecutor")
    rg.return_value.submit.return_value.result.return_value.json.return_value = "a"

    client = ApiClient(in_cluster=False)

    result = client.get_all_metrics()

    assert result is not None
    assert result.result().json() == "a"


def test_get_run_metrics(mocker, kubernetes_api_client_node_port):
    mocker.patch("kubernetes.config.load_kube_config")
    rg = mocker.patch("concurrent.futures.ThreadPoolExecutor")
    rg.return_value.submit.return_value.result.return_value.json.return_value = "a"

    client = ApiClient(in_cluster=False)

    result = client.get_run_metrics("1", since=datetime.datetime.now(), summarize=100)

    assert result is not None
    assert result.result().json() == "a"


def test_get_pod_metrics(mocker, kubernetes_api_client_node_port):
    mocker.patch("kubernetes.config.load_kube_config")
    rg = mocker.patch("concurrent.futures.ThreadPoolExecutor")
    rg.return_value.submit.return_value.result.return_value.json.return_value = "a"

    client = ApiClient(in_cluster=False)

    result = client.get_pod_metrics(
        "rel-mlbench-worker-0", since=datetime.datetime.now(), summarize=100
    )

    assert result is not None
    assert result.result().json() == "a"


def test_post_metrics(mocker, kubernetes_api_client_node_port):
    mocker.patch("kubernetes.config.load_kube_config")
    rg = mocker.patch("concurrent.futures.ThreadPoolExecutor")
    rg.return_value.submit.return_value.result.return_value.json.return_value = "a"

    client = ApiClient(in_cluster=False)

    result = client.post_metric("1", "loss", 10.0, cumulative=False)

    assert result is not None
    assert result.result().json() == "a"


def test_get_runs(mocker, kubernetes_api_client_node_port):
    mocker.patch("kubernetes.config.load_kube_config")
    rg = mocker.patch("concurrent.futures.ThreadPoolExecutor")
    rg.return_value.submit.return_value.result.return_value.json.return_value = "a"

    client = ApiClient(in_cluster=False)

    result = client.get_runs()

    assert result is not None
    assert result.result().json() == "a"


def test_get_run(mocker, kubernetes_api_client_node_port):
    mocker.patch("kubernetes.config.load_kube_config")
    rg = mocker.patch("concurrent.futures.ThreadPoolExecutor")
    rg.return_value.submit.return_value.result.return_value.json.return_value = "a"

    client = ApiClient(in_cluster=False)

    result = client.get_run("1")

    assert result is not None
    assert result.result().json() == "a"


def test_create_run_official(mocker, kubernetes_api_client_node_port):
    mocker.patch("kubernetes.config.load_kube_config")
    rg = mocker.patch("concurrent.futures.ThreadPoolExecutor")
    rg.return_value.submit.return_value.result.return_value.json.return_value = "a"

    client = ApiClient(in_cluster=False)

    result = client.create_run(
        "test_run",
        5,
        num_cpus=4.1,
        max_bandwidth=10000,
        image="PyTorch Cifar-10 ResNet-20",
    )

    assert result is not None
    assert result.result().json() == "a"


def test_create_run_custom(mocker, kubernetes_api_client_node_port):
    mocker.patch("kubernetes.config.load_kube_config")
    rg = mocker.patch("concurrent.futures.ThreadPoolExecutor")
    rg.return_value.submit.return_value.result.return_value.json.return_value = "a"

    client = ApiClient(in_cluster=False)

    result = client.create_run(
        "test_run",
        5,
        num_cpus=4.1,
        max_bandwidth=10000,
        custom_image_name="localhost:5000/mlbench_worker:latest",
        custom_image_command="/.openmpi/bin/mpirun /app/main.py",
        run_all_nodes=False,
    )

    assert result is not None
    assert result.result().json() == "a"


def test_get_worker_pods(mocker, kubernetes_api_client_node_port):
    mocker.patch("kubernetes.config.load_kube_config")
    rg = mocker.patch("concurrent.futures.ThreadPoolExecutor")
    rg.return_value.submit.return_value.result.return_value.json.return_value = "a"

    client = ApiClient(in_cluster=False)

    result = client.get_worker_pods()

    assert result is not None
    assert result.result().json() == "a"



import pytest

import torch

from mlbench_core.models.pytorch.resnet import *
from mlbench_core.models.pytorch.linear_models import *


def test_resnet18():
    resnet = resnet18_bkj(1000)

    inp = torch.rand(2, 3, 32, 32)

    outp = resnet(inp)

    assert outp is not None
    assert outp.shape[0] == 2
    assert outp.shape[1] == 1000

    resnet = resnet18_bkj(500)

    inp = torch.rand(3, 3, 32, 32)

    outp = resnet(inp)

    assert outp is not None
    assert outp.shape[0] == 3
    assert outp.shape[1] == 500


def test_resnet20():
    resnet = get_resnet_model("resnet20", 1, "fp32")

    inp = torch.rand(2, 3, 32, 32)

    outp = resnet(inp)

    assert outp is not None
    assert outp.shape[0] == 2
    assert outp.shape[1] == 10

    resnet = get_resnet_model("resnet20", 1, "fp32")

    inp = torch.rand(3, 3, 32, 32)

    outp = resnet(inp)

    assert outp is not None
    assert outp.shape[0] == 3
    assert outp.shape[1] == 10


def test_resnet20v2():
    resnet = get_resnet_model("resnet20", 2, "fp32")

    inp = torch.rand(2, 3, 32, 32)

    outp = resnet(inp)

    assert outp is not None
    assert outp.shape[0] == 2
    assert outp.shape[1] == 10

    resnet = get_resnet_model("resnet20", 2, "fp32")

    inp = torch.rand(3, 3, 32, 32)

    outp = resnet(inp)

    assert outp is not None
    assert outp.shape[0] == 3
    assert outp.shape[1] == 10


def test_linear_regression():
    lr = LinearRegression(10)  # Linear regression with 10 features
    inp = torch.rand(100, 10)

    output = lr(inp)
    assert output is not None
    assert output.shape[0] == 100
    assert output.shape[1] == 1


def test_logistic_regression():
    log = LogisticRegression(10)

    inp = torch.rand(100, 10)

    output = log(inp)
    assert output is not None
    assert output.shape[0] == 100
    assert output.shape[1] == 1



import torch
import torch.distributed as dist

try:
    import horovod.torch as hvd
except ImportError as e:
    pass


# TODO: those 3 funtions are never used, maybe delete them ?


def broadcast(tensor, src):
    return dist.broadcast(tensor, src=src)


def elementwise_min(tensor):
    dist.all_reduce(tensor, op=dist.ReduceOp.MIN)
    return tensor


def aggregate_gradients(model, world_size, average_models=False):
    """Average gradients of models across all processes."""
    # all_reduce the gradients.
    for ind, param in enumerate(model.parameters()):
        # all reduce.
        dist.all_reduce(param.grad.data, op=dist.ReduceOp.SUM)

        if average_models:
            param.grad.data /= world_size


def global_average(sum, count):
    def helper(array):
        array = get_backend_tensor(torch.Tensor(array))

        dist.all_reduce(array, op=dist.ReduceOp.SUM)
        return array[0] / array[1]

    avg = helper([sum, count])
    return avg


def pack_tensors(tensors, use_cuda=False):
    """
    Packs a list of tensors into one 1-dimensional tensor.

    Args:
        tensors (list[torch.Tensor]): The tensors to pack
        use_cuda (bool): Whether the resulting tensor should be on cuda

    Returns:
        (torch.Tensor, list[int], list[(int, int)]):
            The flattened tensors, the list start indices of each packed tensor,
            and the original shape of each tensor.

            Those values are used to then unpack the tensor
    """
    tensor_sizes = [t.size() for t in tensors]
    pointers = [0]
    for tensor in tensors:
        pointers.append(pointers[-1] + tensor.nelement())

    vec = torch.empty(
        pointers[-1],
        device=tensors[0].device if tensors[0].is_cuda and use_cuda else "cpu",
        dtype=tensors[0].dtype,
    )

    for tensor, start_idx, end_idx in zip(tensors, pointers[:-1], pointers[1:]):
        vec[start_idx:end_idx] = tensor.data.view(-1)

    return vec, pointers, tensor_sizes


def unpack_tensors(aggregated, indices, sizes):
    """
    Unpacks a 1-dimensional tensor into a list of tensors

    Args:
        aggregated (torch.Tensor): The 1-dimensional tensor
        indices (List[Int]): The start index of each tensor
        sizes (List[(Int, Int)]): The size of each resulting tensor

    Returns:
        List[torch.Tensor]: The unpacked tensors
    """
    start_index = indices[:-1]
    end_index = indices[1:]

    tensors = []
    for i, (start, end) in enumerate(zip(start_index, end_index)):
        tensors.append(aggregated[start:end].view(sizes[i]))

    return tensors


##########################################################################################


class Aggregation(object):
    """Aggregate udpates / models from different processes."""

    def __init__(self, use_cuda=False):
        """

        Args:
            use_cuda (bool): Whether to use CUDA tensors for communication
        """
        self.use_cuda = use_cuda

    def _agg(self, data, op):
        """Aggregate data using `op` operation.

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggragated.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.

        Returns:
            :obj:`torch.Tensor`: An aggregated tensor.
        """
        raise NotImplementedError

    def _agg_weights_by_model(self, model, op):
        """Aggregate models by model weight, all layers at once

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
        """
        # Pack all layers
        packed, indices, sizes = pack_tensors(
            [t.data for t in model.parameters()], use_cuda=self.use_cuda
        )
        aggregated = self._agg(packed, op=op)

        tensors = unpack_tensors(aggregated, indices, sizes)
        # Unpack
        for i, param in enumerate(model.parameters()):
            param.data = tensors[i]

    def _agg_gradients_by_model(self, model, op):
        """Aggregate models gradients, all layers at once

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
        """
        # Pack all layers

        packed, indices, sizes = pack_tensors(
            [t.grad.data for t in model.parameters()], use_cuda=self.use_cuda
        )
        aggregated = self._agg(packed, op=op)

        # Unpack
        tensors = unpack_tensors(aggregated, indices, sizes)
        for i, param in enumerate(model.parameters()):
            param.grad.data = tensors[i]

    def _agg_weights_by_layer(self, model, op):
        """Aggregate models by model weight, for each layer individually

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
        """
        # Aggregate layer by layer
        for _, param in enumerate(model.parameters()):
            grad = self._agg(param.data, op=op)
            param.data = grad

    def _agg_gradients_by_layer(self, model, op):
        """Aggregate models gradients each layer individually

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
        """
        # Aggregate layer by layer
        for _, param in enumerate(model.parameters()):
            grad = self._agg(param.grad.data, op=op)
            param.grad.data = grad

    def agg_model(self, by_layer=False):
        if by_layer:
            return self._agg_weights_by_layer
        else:
            return self._agg_weights_by_model

    def agg_grad(self, by_layer=False):
        if by_layer:
            return self._agg_gradients_by_layer
        else:
            return self._agg_gradients_by_model


class AllReduceAggregation(Aggregation):
    """Aggregate udpates / models from different processes."""

    def __init__(self, world_size, use_cuda=False):
        self.world_size = world_size
        super(AllReduceAggregation, self).__init__(use_cuda=use_cuda)

    def _agg(self, data, op):
        """Aggregate data using `op` operation.

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggragated.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.

        Returns:
            :obj:`torch.Tensor`: An aggregated tensor.
        """
        if op == "avg":
            dist.all_reduce(data, op=dist.ReduceOp.SUM)
            data /= self.world_size
        else:
            raise NotImplementedError
        return data


class AllReduceAggregationFP16(AllReduceAggregation):
    def _agg(self, data, op):
        """Aggregate data using `op` operation.
        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggregated (Should be `torch.float16`)
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
        Returns:
            :obj:`torch.Tensor`: An aggregated tensor.
        """
        if op == "avg":
            data = hvd.allreduce(data, op=hvd.Sum) / self.world_size
        else:
            raise NotImplementedError
        return data


class DecentralizedAggregation(Aggregation):
    """Aggregate updates in a decentralized manner."""

    def __init__(self, rank, neighbors, use_cuda=False):
        """
        Args:
            rank (int): Rank of the current process
            neighbors (list): A list of ranks of its neighbors.
        """
        assert rank not in neighbors
        self.rank = rank
        self.neighbors = neighbors
        super(DecentralizedAggregation, self).__init__(use_cuda=use_cuda)

    def _agg(self, data, op):
        """Aggregate data using `op` operation.

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggragated.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.

        Returns:
            :obj:`torch.Tensor`: An aggregated tensor.
        """
        # Create some tensors to host the values from neighborhood.
        local_data = {i: torch.zeros_like(data) for i in self.neighbors}
        local_data[self.rank] = data

        reqs = []
        for node in self.neighbors:
            reqs.append(dist.isend(tensor=local_data[self.rank], dst=node))
            reqs.append(dist.irecv(tensor=local_data[node], src=node))

        for req in reqs:
            req.wait()

        # Aggregate local_data
        if op == "avg":
            output = sum(local_data.values()) / (len(self.neighbors) + 1)
        else:
            raise NotImplementedError("op {} is not supported yet.".format(op))

        return output


class SparsifiedAggregation(Aggregation):
    """Aggregate sparsified updates."""

    def __init__(self, model, use_cuda=False):
        super(SparsifiedAggregation, self).__init__(use_cuda=use_cuda)
        pass

    def _agg(self, data, op):
        pass


def get_backend_tensor(tensor):
    if dist.is_initialized() and dist.get_backend() == dist.Backend.NCCL:
        return tensor.cuda()
    return tensor



import os

import torch
from mlbench_core.dataset.translation.pytorch import config
from mlbench_core.dataset.translation.pytorch.tokenizer import WMT14Tokenizer
from mlbench_core.dataset.util.tools import maybe_download_and_extract_tar_gz
from torchtext.data import Dataset


def _construct_filter_pred(min_len, max_len):
    """
    Constructs a filter predicate
    Args:
        min_len (int): Min sentence length
        max_len (int): Max sentence length

    Returns:
        func
    """
    filter_pred = lambda x: not (x[0] < min_len or x[1] < min_len)
    if max_len is not None:
        filter_pred = lambda x: not (
            x[0] < min_len or x[0] > max_len or x[1] < min_len or x[1] > max_len
        )

    return filter_pred


def build_collate_fn(batch_first, sort):
    def collate_seq(seq):
        """
        Builds batches for training or inference.
        Batches are returned as pytorch tensors, with padding.

        :param seq: list of sequences
        """
        lengths = torch.tensor([len(s) for s in seq], dtype=torch.int64)
        batch_length = max(lengths)

        shape = (len(seq), batch_length)
        seq_tensor = torch.full(shape, config.PAD, dtype=torch.int64)

        for i, s in enumerate(seq):
            end_seq = lengths[i]
            segmented = torch.Tensor(s[:end_seq])
            seq_tensor[i, :end_seq].copy_(segmented)

        if not batch_first:
            seq_tensor = seq_tensor.t()

        return seq_tensor, lengths

    def parallel_collate(seqs):
        """
        Builds batches from parallel dataset (src, tgt), optionally sorts batch
        by src sequence length.

        :param seqs: tuple of (src, tgt) sequences
        """
        src_seqs, trg_seqs = zip(*seqs)
        if sort:
            indices, src_seqs = zip(
                *sorted(enumerate(src_seqs), key=lambda x: len(x[1]), reverse=True)
            )
            trg_seqs = [trg_seqs[idx] for idx in indices]

        src_seqs = collate_seq(src_seqs)
        trg_seqs = collate_seq(trg_seqs)
        return src_seqs, trg_seqs

    return parallel_collate


class WMT14Dataset(Dataset):
    """Dataset for WMT14 en to de translation
    Based on `torchtext.datasets.WMT14`

    Args:
        root (str): Root folder where to download files
        lang (dict): Language translation pair
        math_precision (str): One of `fp16` or `fp32`. The precision used during training
        download (bool): Download the dataset from source
        train (bool): Load train set
        validation (bool): Load validation set
        lazy (bool): Load the dataset in a lazy format
        min_len (int): Minimum sentence length
        max_len (int | None): Maximum sentence length
        max_size (int | None): Maximum dataset size
    """

    urls = [
        (
            "https://storage.googleapis.com/mlbench-datasets/translation/wmt16_en_de.tar.gz",
            "wmt16_en_de.tar.gz",
        )
    ]
    name = "wmt14"
    dirname = ""

    def __init__(
        self,
        root,
        lang=None,
        math_precision=None,
        download=True,
        train=False,
        validation=False,
        lazy=False,
        sort=False,
        min_len=0,
        max_len=None,
        max_size=None,
    ):
        self.lazy = lazy

        super(WMT14Dataset, self).__init__(examples=[], fields={})
        if download:
            url, file_name = self.urls[0]
            maybe_download_and_extract_tar_gz(root, file_name, url)

        src_tokenizer = WMT14Tokenizer(root, lang=lang, math_precision=math_precision,)
        trg_tokenizer = WMT14Tokenizer(root, lang=lang, math_precision=math_precision,)

        self.vocab_size = src_tokenizer.vocab_size
        self.fields = {"src": src_tokenizer, "trg": trg_tokenizer}

        self.max_len = max_len
        self.min_len = min_len

        if train:
            path = os.path.join(root, config.TRAIN_FNAME)
        elif validation:
            path = os.path.join(root, config.VAL_FNAME)
        else:
            raise NotImplementedError()

        self.examples, self.indices = self._process_data(
            path,
            filter_pred=_construct_filter_pred(min_len, max_len),
            lazy=lazy,
            sort=sort,
            max_size=max_size,
        )

    def _process_data(self, path, filter_pred, sort=False, lazy=False, max_size=None):
        """Loads data from given path and processes the lines

        Args:
            path (str): Dataset directory path
            filter_pred (func): Filter predicate function (to filter inputs)
            lazy (bool): Whether to load the dataset in lazy mode
            max_size (int | None): Maximum size of dataset

        Returns:
            List: The list of examples
        """
        src_path, trg_path = tuple(os.path.expanduser(path + x) for x in config.EXTS)
        examples = []
        src_lengths = []
        with open(src_path, mode="r", encoding="utf-8") as src_file, open(
            trg_path, mode="r", encoding="utf-8"
        ) as trg_file:
            for src_line, trg_line in zip(src_file, trg_file):
                src_line, trg_line = src_line.strip(), trg_line.strip()

                should_consider = filter_pred(
                    (src_line.count(" ") + 1, trg_line.count(" ") + 1)
                )
                if src_line != "" and trg_line != "" and should_consider:
                    src_lengths.append(src_line.count(" ") + 1)
                    if lazy:
                        examples.append((src_line, trg_line))
                    else:
                        examples.append(self._parse_example((src_line, trg_line)))

                if max_size and len(examples) >= max_size:
                    break

        indices = list(range(len(examples)))
        if sort:
            indices, _ = zip(*sorted(enumerate(src_lengths), key=lambda x: x[1]))
        return examples, indices

    def __len__(self):
        return len(self.examples)

    def _parse_example(self, example):
        src_line, trg_line = example
        return (
            self.fields["src"].parse_line(src_line),
            self.fields["trg"].parse_line(trg_line),
        )

    def __getitem__(self, item):
        if self.lazy:
            return self._parse_example(self.examples[self.indices[item]])
        else:
            return self.examples[self.indices[item]]



from torch.nn import init as init


def init_lstm_(lstm, init_weight=0.1):
    """
    Initializes weights of LSTM layer.
    Weights and biases are initialized with uniform(-init_weight, init_weight)
    distribution.

    Args:
        lstm (torch.nn.LSTM):
        init_weight (float): range for the uniform initializer

    """
    # Initialize hidden-hidden weights
    init.uniform_(lstm.weight_hh_l0.data, -init_weight, init_weight)
    # Initialize input-hidden weights:
    init.uniform_(lstm.weight_ih_l0.data, -init_weight, init_weight)

    # Initialize bias. PyTorch LSTM has two biases, one for input-hidden GEMM
    # and the other for hidden-hidden GEMM. Here input-hidden bias is
    # initialized with uniform distribution and hidden-hidden bias is
    # initialized with zeros.
    init.uniform_(lstm.bias_ih_l0.data, -init_weight, init_weight)
    init.zeros_(lstm.bias_hh_l0.data)

    if lstm.bidirectional:
        init.uniform_(lstm.weight_hh_l0_reverse.data, -init_weight, init_weight)
        init.uniform_(lstm.weight_ih_l0_reverse.data, -init_weight, init_weight)

        init.uniform_(lstm.bias_ih_l0_reverse.data, -init_weight, init_weight)
        init.zeros_(lstm.bias_hh_l0_reverse.data)



import enum
import json
import os
import shutil
import dill
import torch


class CheckpointFreq(enum.IntEnum):
    # Checkpoint every epoch or best epoch
    ALL = 1
    BEST = 2
    NONE = 3


class Checkpointer(object):
    """ A class for handling checkpoint saving and loading.

    Args:
        ckpt_run_dir (str): The path of the checkpoint directory.
        rank (int): The rank of the eurrent worker.
        freq (int): The frequency of checkpointing. Default: `CheckpointFreq.BEST`
        save_stats (bool): Save stats to additional text files. Default: `True`
    """

    def __init__(self, ckpt_run_dir, rank, freq=CheckpointFreq.BEST, save_stats=True):
        self.dirname = ckpt_run_dir
        self.rank = rank
        self.freq = freq
        self.save_stats = save_stats
        # self.runtime = {'cumu_time_val': []}

    def save(self, tracker, model, optimizer, scheduler, epoch, is_best):
        """ Saves a checkpoint

        Args:
            tracker (:obj:`mlbench_core.utils.pytorch.helpers.Tracker`): The
                metrics tracker object
            model (:obj:`torch.nn.Module`): a pytorch model to be trained and validated.
            optimizer (:obj:`torch.optim.Optimizer`): an optimizer for the given model.
            scheduler (:obj:`mlbench_core.lr_scheduler.pytorch.lr.*`): a scheduler for
                hyperparameters.
            epoch (int): The current epoch
            is_best (bool): Whether the current model is a new best scoring one
        """
        state = {
            "tracker": tracker,
            "model": model.state_dict(),
            "optimizer": optimizer.state_dict(),
            "scheduler": scheduler.state_dict(),
            "freq": self.freq,
        }

        filename = "{epoch}_{rank}.pth.tar".format(
            epoch=tracker.current_epoch, rank=self.rank
        )
        checkpoint_path = os.path.join(self.dirname, filename)
        best_model_path = os.path.join(self.dirname, "model_best.pth.tar")

        if self.freq == CheckpointFreq.ALL:
            torch.save(state, checkpoint_path, pickle_module=dill)
            if is_best:
                shutil.copyfile(checkpoint_path, best_model_path)
        elif self.freq == CheckpointFreq.BEST:
            torch.save(state, best_model_path, pickle_module=dill)
        elif self.freq != CheckpointFreq.NONE:
            raise NotImplementedError

        self._maybe_save_stats(tracker.records, tracker.current_epoch, self.rank)

    def _maybe_save_stats(self, records, epoch, rank):
        """Save the records in the tracker."""
        if self.save_stats:
            filename = os.path.join(self.dirname, "{}_{}.json".format(epoch, rank))
            with open(filename, "w") as f:
                json.dump(records, f)

    @staticmethod
    def load(ckpt_run_dir, rank, model, optimizer, scheduler):
        """ Loads a checkpoint

        Args:
            ckpt_run_dir (str): Folder path of checkpoint directory
            rank (int): The rank of the current worker
            model (:obj:`torch.nn.Module`): a pytorch model to be trained and validated.
            optimizer (:obj:`torch.optim.Optimizer`): an optimizer for the given model.
            scheduler (:obj:`mlbench_core.lr_scheduler.pytorch.lr.*`): a scheduler for
                hyperparameters.

        Returns:
            A tuple of `(Checkpointer, model, optimizer, scheduler)`
        """

        checkpoint_path = determine_restore_ckpt_path(rank, ckpt_run_dir)

        if not os.path.isfile(checkpoint_path):
            raise FileNotFoundError(
                "No checkpoint found at '{}' for rank '{}'".format(ckpt_run_dir, rank)
            )

        checkpoint = torch.load(checkpoint_path, pickle_module=dill)

        model.load_state_dict(checkpoint["model"])
        optimizer.load_state_dict(checkpoint["optimizer"])
        scheduler.load_state_dict(checkpoint["scheduler"])

        tracker = checkpoint["tracker"]

        freq = checkpoint["freq"]

        checkpointer = Checkpointer(ckpt_run_dir, rank, freq)
        # checkpointer.runtime['cumu_time_val'] = checkpoint['tracker']['cumu_time_val']

        return checkpointer, model, optimizer, scheduler, tracker

    @staticmethod
    def load_model_by_epoch(ckpt_run_dir, rank, epoch, model):
        """ Loads a checkpoint

        Args:
            ckpt_run_dir (str): Folder path of checkpoint directory
            rank (int): The rank of the current worker
            epoch (int): Epoch of the model to be loaded.
            model (:obj:`torch.nn.Module`): a pytorch model to be trained and validated.

        Returns:
            `model`
        """
        checkpoint_path = os.path.join(
            ckpt_run_dir, "{}_{}.pth.tar".format(epoch, rank)
        )

        if not os.path.isfile(checkpoint_path):
            raise FileNotFoundError(
                "No checkpoint found at '{}' for rank '{}'".format(ckpt_run_dir, rank)
            )

        checkpoint = torch.load(checkpoint_path)

        model.load_state_dict(checkpoint["model"])
        return model

    @staticmethod
    def checkpoint_exists(ckpt_run_dir, rank, epoch):
        """ Check if a checkpoint exists.

        Args:
            ckpt_run_dir (str): Folder path of checkpoint directory
            rank (int): The rank of the current worker
            epoch (int): Epoch of the model to be loaded.

        Returns:
            `model`
        """
        checkpoint_path = os.path.join(
            ckpt_run_dir, "{}_{}.pth.tar".format(epoch, rank)
        )
        if not os.path.isfile(checkpoint_path):
            raise FileNotFoundError(
                "No checkpoint found at '{}' for rank '{}'".format(ckpt_run_dir, rank)
            )


def determine_restore_ckpt_path(rank, checkpoint_root):
    """Determine the checkpoint path to restore.

    Args:
        rank (int): The rank of the current worker
        checkpoint_root (str): Folder path of checkpoint directory

    Returns:
        The path of the newest checkpoint for this worker
    """
    ckpt_ids = os.listdir(checkpoint_root)
    ckpt_ids = list(filter(lambda x: x.endswith(".pth.tar"), ckpt_ids))
    ckpt_ids = list(set(ckpt_ids) - set(["model_best.pth.tar"]))

    ckpt_ids = filter(
        lambda x: x.split("_")[1][: -len(".pth.tar")] == str(rank), ckpt_ids
    )

    latest = sorted(ckpt_ids, reverse=True, key=lambda x: int(x.split("_")[0]))

    path = os.path.join(checkpoint_root, latest[0])
    return path



import torch


class LogisticRegression(torch.nn.Module):
    """
    Logistic regression implementation

    Args:
        n_features (int): Number of features
    """

    def __init__(self, n_features):
        super(LogisticRegression, self).__init__()

        self.linear = torch.nn.Linear(n_features, 1, bias=False)

    def forward(self, x):
        y_pred = torch.sigmoid(self.linear(x))
        return y_pred


class LinearRegression(torch.nn.Module):
    """
    Ridge regression implementation

    Args:
        n_features (int): Number of features
    """

    def __init__(self, n_features):
        super(LinearRegression, self).__init__()
        self.linear = torch.nn.Linear(n_features, 1, bias=False)

    def forward(self, x):
        return self.linear(x)



"""Initialize environment for pytorch."""

import tensorflow as tf


def _init_cleanup(config):
    r"""Cleanup legacy files like logs, output."""
    print("=> Initial cleanup")


def _init_log(config):
    print("=> Initialize log")


def _init_tensorflow(config):
    print("=> Initialize TensorFlow")


def initialize_backends(config):
    """Initializes the backends.

    Sets up logging, sets up tensorflow and configures paths
    correctly.

    Args:
        config (:obj:`types.SimpleNamespace`): a global object containing all of the config.

    Returns:
        (:obj:`types.SimpleNamespace`): a global object containing all of the config.
    """
    _init_cleanup(config)

    _init_log(config)

    _init_tensorflow(config)
    return config


def default_session_config(
    tf_allow_soft_placement, tf_log_device_placement, tf_gpu_mem
):
    """Initialize session configuration."""
    session_conf = tf.ConfigProto(
        allow_soft_placement=tf_allow_soft_placement,
        log_device_placement=tf_log_device_placement,
    )

    session_conf.gpu_options.allow_growth = False  # True
    session_conf.gpu_options.per_process_gpu_memory_fraction = tf_gpu_mem
    return session_conf



import logging

import torch
import torch.optim
import torch.utils.data
from mlbench_core.utils import AverageMeter
from mlbench_core.utils.pytorch.distributed import global_average
from datetime import datetime
logger = logging.getLogger("mlbench")
LOG_EVERY_N_BATCHES = 25


def _record_train_batch_stats(
    batch_idx, loss, batch_size, tracker, num_batches_per_device_train
):
    """Record the stats in a training batch.

    Args:
        batch_idx (int): The id of the current batch
        loss (float): The loss of the batch
        tracker (`obj`:mlbench_core.utils.Tracker): Tracker object to use.
        num_batches_per_device_train (int): Number of batches per train epoch
    """
    progress = batch_idx / num_batches_per_device_train
    progress += tracker.current_epoch

    log_to_api = (
        batch_idx % LOG_EVERY_N_BATCHES == 0
        or batch_idx == num_batches_per_device_train
    )

    if tracker:
        tracker.record_loss(loss, batch_size, log_to_api=log_to_api)
    status = "Epoch {:5.2f} Batch {:4}: ".format(progress, batch_idx)

    logger.info(status + str(tracker))


# def _iterate_loader(dataloader, use_cuda=False):
#     for (data, target) in dataloader:
#         if use_cuda:
#             data = data[0].cuda(), data[1].cuda()
#             target = target[0].cuda(), target[1]
#
#         yield data, target

def _maybe_cuda(example, use_cuda=False):
    src, trg = example
    if use_cuda:
        src = src[0].cuda(), src[1].cuda()
        trg = trg[0].cuda(), trg[1].cuda()
    return src, trg


class GNMTTrainer:
    """ Trainer used for GNMT model
    Args:
        model (`obj`:mlbench_core.models.pytorch.gnmt.GNMT): The GNMT model
        criterion (`obj`:torch.nn.Module): The loss function to use
        fp_optimizer: Floating point optimizer (either for fp16 or fp32)
        scheduler (`obj`:torch.optim.LambdaLR): Learning Rate scheduler
        translator (`obj`:mlbench_core.evaluation.pytorch.inference.Translator): Translator class
        rank (int): Current node rank
        schedule_per (str): Scheduler per `batch` or `epoch`
        tracker (`obj`:mlbench_core.utils.Tracker): Tracker object to use.
        metrics (list): The list of metrics to use
        iter_size (int): Number of iterations to do before calling `optimizer.step`

    """

    def __init__(
        self,
        model,
        criterion,
        fp_optimizer,
        scheduler,
        translator,
        rank,
        schedule_per,
        tracker,
        metrics,
        iter_size,
        use_cuda,
    ):
        self.model = model
        self.batch_first = model.batch_first
        self.criterion = criterion
        self.epoch = 0
        self.rank = rank
        # Optimizers & Scheduler
        self.fp_optimizer = fp_optimizer

        self.schedule_per = schedule_per
        self.scheduler = scheduler
        self.device = next(model.parameters()).device

        self.translator = translator
        self.metrics = metrics
        self.iter_size = iter_size

        self.tracker = tracker

        self.use_cuda = use_cuda

    def compute_model_output(self, src, trg):
        """ Computes output of GNMT model

        Args:
            src (tuple): Source data point. Should be tuple of (tokens, lengths)
            trg (tuple): Target data point. Should be tuple of (tokens, lengths)

        Returns:
            `obj`:torch.Tensor: The output tensor
        """
        src, src_length = src
        trg, _ = trg

        if self.batch_first:
            output = self.model(src, src_length, trg[:, :-1])
        else:
            output = self.model(src, src_length, trg[:-1])

        return output

    def set_tracker(self, tracker):
        """ Sets the trainer's tracker

        Args:
            tracker (`obj`:mlbench_core.utils.Tracker): Tracker object to use.
        """
        self.tracker = tracker

    def compute_loss(self, src, trg, output):
        """ Computes the Loss of a given input and output

        Args:
            src (tuple): Source data point. Should be tuple of (tokens, lengths)
            trg (tuple): Target data point. Should be tuple of (tokens, lengths)
            output (`obj`:torch.Tensor): Output of given input

        Returns:
            (float, float, float, dict): Total loss, loss per token, loss per sentence, num tokens
        """
        src, src_length = src
        trg, trg_length = trg

        num_toks = {"trg": int(sum(trg_length - 1)), "src": int(sum(src_length))}

        if self.batch_first:
            tgt_labels = trg[:, 1:]
            T, B = output.size(1), output.size(0)

        else:
            tgt_labels = trg[1:]
            T, B = output.size(0), output.size(1)

        loss = self.criterion(output.view(T * B, -1), tgt_labels.contiguous().view(-1))

        loss_per_batch = loss.item()
        loss /= B * self.iter_size

        loss_per_token = loss_per_batch / num_toks["trg"]
        loss_per_sentence = loss_per_batch / B

        return loss, loss_per_token, loss_per_sentence, num_toks

    def optimize(self, batch_idx, data, num_batches_per_device_train):
        """ Optimizes on a given batch

        Args:
            batch_idx (int): Index of the batch
            data (`obj`:torch.Tensor): Batch tensor
            num_batches_per_device_train (int): Number of batches per train epoch
        """

        # Whether to update the weights at this iteration
        update = (batch_idx % self.iter_size) == self.iter_size - 1
        if self.tracker:
            self.tracker.batch_start()

        # Clear gradients in the optimizer.
        if (batch_idx % self.iter_size) == 0:
            self.fp_optimizer.zero_grad()
            if self.tracker:
                self.tracker.record_batch_step("init")

        # Compute the output
        src, trg = data
        output = self.compute_model_output(src, trg)

        if self.tracker:
            self.tracker.record_batch_step("fwd_pass")

        # Compute the loss
        stats = self.compute_loss(src, trg, output)
        loss, loss_per_token, loss_per_sentence, num_toks = stats
        #

        if self.tracker:
            self.tracker.record_batch_step("comp_loss")

        # Backprop
        self.fp_optimizer.backward_loss(loss)

        if self.tracker:
            self.tracker.record_batch_step("backprop")

        # Opt step
        if update:
            self.fp_optimizer.step()
            if self.tracker:
                self.tracker.record_batch_step("opt_step")

            # Learning rate scheduler
            if self.schedule_per == "batch":
                self.scheduler.step()

        if self.tracker:
            self.tracker.batch_end()

        if self.batch_first:
            batch_size = output.size(0)
        else:
            batch_size = output.size(1)

        _record_train_batch_stats(
            batch_idx=batch_idx,
            loss=loss_per_token,
            batch_size=batch_size,
            tracker=self.tracker,
            num_batches_per_device_train=num_batches_per_device_train,
        )

    def _training(self):
        """Sets the model and tracker in training"""
        self.model.train()

        if self.tracker:
            self.tracker.train()

    def _eval(self):
        """Sets the model and tracker in evaluation mode"""
        self.model.eval()

        # Set tracker in validation mode
        if self.tracker:
            self.tracker.validation()
            self.tracker.validation_start()

    def train_round(
        self, train_loader, val_loader, bleu_score=False, validate_every=None
    ):
        """ Performs one epoch of training

        Args:
            train_loader: The train set loader
            val_loader: The validation set loader
            bleu_score (bool): Compute bleu score during epoch
            validate_every (int | None): Validate every n batches.
                Default `None` (no validation)
        """
        # Set in training mode
        self._training()

        num_batches_per_device_train = len(train_loader)

        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        # data_iter = _iterate_loader(train_loader, use_cuda=self.use_cuda)
        start = datetime.now()
        for batch_idx, data in enumerate(train_loader):
            end = datetime.now()
            logger.info("Batch {} took {}".format(batch_idx, end - start))
            data = _maybe_cuda(data, use_cuda=self.use_cuda)
            self.optimize(batch_idx, data, num_batches_per_device_train)

            start = datetime.now()
            if bleu_score and (batch_idx + 1) % validate_every == 0:
                self.validation_round(val_loader)
                self._training()

        if self.schedule_per == "epoch":
            self.scheduler.step()

    def validate(self, loader):
        """Performs validation of the validation set

        Args:
            loader: The data set loader

        Returns:
            (dict, float): The metrics averages and the loss average
        """
        losses = AverageMeter()

        # Reset metrics
        for metric in self.metrics:
            metric.reset()

        with torch.no_grad():
            # data_iter = _iterate_loader(loader, use_cuda=self.use_cuda)
            for data in loader:
                src, trg = _maybe_cuda(data, use_cuda=self.use_cuda)
                output = self.compute_model_output(src, trg)

                # Compute loss
                stats = self.compute_loss(src, trg, output)
                loss, loss_per_token, loss_per_sentence, num_toks = stats

                # Update loss
                losses.update(loss_per_token, num_toks["trg"])

                # Update metrics
                translated, targets = self.translator.translate(src, trg)
                for metric in self.metrics:
                    metric_value = metric(loss, translated, targets)
                    size = src[0].shape[0] if self.batch_first else src[0].shape[1]

                    metric.update(metric_value, size)

        metrics_averages = {metric: metric.average().item() for metric in self.metrics}
        loss_average = global_average(losses.sum, losses.count).item()
        return metrics_averages, loss_average

    def validation_round(self, data_loader):
        """Performs one validation round and checks if the goal was reached

        Args:
            data_loader: Data loader

        Returns:
            (bool): Whether this validation is the best so far
        """
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        tracker = self.tracker
        self._eval()
        # Gather metrics and loss average
        metrics_values, loss = self.validate(data_loader)
        if tracker:
            tracker.validation_end()

        if len(metrics_values) > 0:
            # Save metrics
            if tracker:
                for metric, value in metrics_values.items():
                    tracker.record_metric(metric, value, log_to_api=True)

                    global_metric_value = global_average(value, 1).item()

                    if self.rank == 0:
                        tracker.record_stat(
                            "global_{}".format(metric.name),
                            global_metric_value,
                            log_to_api=True,
                        )

            #
            if self.rank == 0 and tracker:
                logger.info(
                    "{} for rank {}:(best epoch {}, current epoch {}): {:.3f}".format(
                        tracker.primary_metric.name,
                        tracker.rank,
                        tracker.best_epoch,
                        tracker.current_epoch,
                        tracker.best_metric_value,
                    )
                )
        else:
            if self.rank == 0:
                logger.info("Validation loss={:.3f}".format(loss))

        if tracker:
            tracker.record_loss(loss, log_to_api=True)

            global_loss = global_average(loss, 1).item()

            if self.rank == 0:
                tracker.record_stat("global_loss", global_loss, log_to_api=True)

        return tracker.is_best() if tracker else False



r""" Create dataset and dataloader in PyTorch. """
import logging
import os

import spacy
from spacy.symbols import ORTH

import torchtext
import torchtext.datasets as nlp_datasets


def _get_text():
    spacy_en = spacy.load("en")
    spacy_en.tokenizer.add_special_case("<eos>", [{ORTH: "<eos>"}])
    spacy_en.tokenizer.add_special_case("<bos>", [{ORTH: "<bos>"}])
    spacy_en.tokenizer.add_special_case("<unk>", [{ORTH: "<unk>"}])

    def spacy_tok(text):
        return [tok.text for tok in spacy_en.tokenizer(text)]

    TEXT = torchtext.data.Field(lower=True, tokenize=spacy_tok, batch_first=False)
    return TEXT


class Wikitext2(nlp_datasets.WikiText2):
    """Wikitext2 Dataset.

    Loads Wikitext2 dataset.
    Based on `torchtext.datasets.Wikitext2`

    Args:
        root (str): Root folder of Imagenet dataset (without `train/` or `val/`)
        train (bool): Whether to get the train or validation set (default=True)
    """

    def __init__(self, root, text_field=None, download=True, train=True):
        self.train = train

        self.text_field = text_field
        if not self.text_field:
            self.text_field = _get_text()

        self.root = root

        if download:
            path = self.download(root)
        else:
            path = os.path.join(root, "wikitext-2/wikitext-2")

        if train:
            path = os.path.join(path, "wiki.train.tokens")
        else:
            path = os.path.join(path, "wiki.valid.tokens")

        super(Wikitext2, self).__init__(path, self.text_field)



""" MLBench Master/Dashboard API Client Functionality """

import concurrent.futures
import datetime
import logging
from kubernetes import client, config
import requests

MLBENCH_BACKENDS = ["MPI", "GLOO", "NCCL"]

MLBENCH_IMAGES = {
    "PyTorch Cifar-10 ResNet-20": (
        "mlbench/pytorch-cifar10-resnet:latest",
        "/conda/bin/python /codes/main.py --run_id {run_id} --rank {rank} --hosts {hosts} --backend {backend}",
        True,
    ),
    "PyTorch Cifar-10 ResNet-20 (Scaling LR)": (
        "mlbench/pytorch-cifar10-resnet-scaling:latest",
        "/conda/bin/python /codes/main.py --run_id {run_id} --rank {rank} --hosts {hosts} --backend {backend}",
        True,
    ),
    "PyTorch Linear Logistic Regression": (
        "mlbench/pytorch-epsilon-logistic-regression-all-reduce:latest",
        "/conda/bin/python /codes/main.py --run_id {run_id} --rank {rank} --hosts {hosts} --backend {backend}",
        True,
    ),
    "PyTorch Machine Translation GNMT": (
        "mlbench/pytorch-wmt14-gnmt-all-reduce:latest",
        "/conda/bin/python /codes/main.py --run_id {run_id} --rank {rank} --hosts {hosts} --backend {backend}",
        True,
    ),
    "Tensorflow Cifar-10 ResNet-20 Open-MPI": (
        "mlbench/tensorflow-cifar10-resnet:latest",
        "/conda/bin/python /codes/main.py --run_id {run_id} --rank {rank} --hosts {hosts} --backend {backend}",
        False,
    ),
}

r"""
Dict of official benchmark images

Note:
    Format: ``{name: (image_name, command, run_on_all, GPU_supported)}``"""


class _CustomApiClient(client.ApiClient):
    """
    Deals with a bug introduced by a fix in swagger.
    https://github.com/kubernetes-client/python/issues/411
    https://github.com/swagger-api/swagger-codegen/issues/6392
    """

    def __del__(self):
        pass


class ApiClient(object):
    """ Client for the mlbench Master/Dashboard REST API

    When used inside a cluster, will use the API Pod IP directly
    for communication. When used outside of a cluster, will try to
    figure out how to access the API depending on the K8s service
    type, if it's accessible. Endpoint URL can also be set manually.

    All requests are executed in a separate process to ensure
    non-blocking execution. Results are returned as
    ``concurrent.futures.Future`` objects wrapping ``requests``
    responses.

    Expects K8s credentials to be set correctly (automatic inside
    a cluster, through kubectl outside of it)

    Args:
        max_workers (int): maximum number of processes to run in parallel
        in_cluster (bool): Whether the client is run inside the K8s cluster
            or not
        label_selector (str): K8s label selectors to find the master pod
            when running inside a cluster. Default:
            ``component=master,app=mlbench``
        k8s_namespace (str): K8s namespace mlbench is running in.
            Default: ``default``
        service_name (str): Name of the master service, usually something
            like ``release-mlbench-master``. Only needed when running
            outside of a cluster. Default: ``None``
        url (str): ip:port/path or hostname:port/path that overrides
            automatic endpoint detection, pointing to the root of the
            master/dashboard node. Default: ``None``"""

    def __init__(
        self,
        max_workers=5,
        in_cluster=True,
        label_selector="component=master,app=mlbench",
        k8s_namespace="default",
        url=None,
        load_config=True,
    ):
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
        self.in_cluster = in_cluster

        self.port = None

        if url is None:
            if self.in_cluster:
                url = self.__get_in_cluster_url(
                    label_selector, k8s_namespace, load_config
                )
            else:
                url = self.__get_out_of_cluster_url(
                    label_selector, k8s_namespace, load_config
                )

        self.endpoint = "http://{url}/api/".format(url=url)

    def __get_in_cluster_url(self, label_selector, k8s_namespace, load_config):
        """Get the API url for the dashboard when running in a cluster """
        if load_config:
            config.load_incluster_config()

        configuration = client.Configuration()
        k8s_client = client.CoreV1Api(_CustomApiClient(configuration))
        namespaced_pods = k8s_client.list_namespaced_pod(
            k8s_namespace, label_selector=label_selector
        )

        assert len(namespaced_pods.items) == 1

        master_pod = namespaced_pods.items[0]

        return master_pod.status.pod_ip + ":80"

    def __get_out_of_cluster_url(self, label_selector, k8s_namespace, load_config):
        """Get the API url for the dashboard when running out of a cluster """
        if load_config:
            config.load_kube_config()

        configuration = client.Configuration()
        k8s_client = client.CoreV1Api(_CustomApiClient(configuration))
        ret = k8s_client.list_service_for_all_namespaces(label_selector=label_selector)

        if len(ret.items) == 0:
            raise ValueError("Couldn't find a deployed dashboard service")

        if len(ret.items) > 1:
            raise ValueError("Found multiple deployed dashboard services")

        service = ret.items[0]
        service_type = service.spec.type

        if service_type == "ClusterIP":
            # cluster ip: up to user to grant access
            ip = service.spec.cluster_ip
            self.port = service.spec.ports[0].port
        elif service_type == "NodePort":
            self.port = next(p.node_port for p in service.spec.ports if p.port == 80)

            if service.spec.external_i_ps and len(service.spec.external_i_ps) > 0:

                ip = service.spec.external_i_ps[0]
            else:
                # try and get public node ip
                namespaced_pods = k8s_client.list_namespaced_pod(
                    k8s_namespace, label_selector=label_selector
                )

                assert len(namespaced_pods.items) == 1

                master_pod = namespaced_pods.items[0]
                node_name = master_pod.spec.node_name
                node = k8s_client.read_node(node_name)
                ip = None
                internal_ip = None
                for address in node.status.addresses:
                    if address.type == "ExternalIP":
                        ip = address.address
                        break
                    if address.type == "InternalIP":
                        internal_ip = address.address
                if not ip:
                    # no external IP found, maybe internal works
                    logging.warning(
                        "No ExternalIP found for NodePort "
                        "Service. Trying InternalIP. This only works "
                        "if the cluster internal IP's are reachable."
                    )
                    ip = internal_ip
        elif service_type == "LoadBalancer":
            self.port = service.spec.ports[0].port

            if service.status.load_balancer.ingress is None:
                raise NotImplementedError(
                    "Service Type 'LoadBalancer' only works with an "
                    "'Ingress' ip defined"
                )
            ip = service.status.load_balancer.ingress.ip
        else:
            raise NotImplementedError

        return "{ip}:{port}".format(ip=ip, port=self.port)

    def get_all_metrics(self):
        """ Get all metrics ever recorded by the master node.

        Returns:
            A ``concurrent.futures.Future`` objects wrapping
            ``requests.response`` object. Get the result by calling
            ``return_value.result().json()``
        """
        request_url = "{endpoint}metrics/".format(endpoint=self.endpoint)
        future = self.executor.submit(requests.get, request_url)

        return future

    def get_run_metrics(
        self, run_id, since=None, summarize=None, metric_filter=None, last_n=None
    ):
        """ Get all metrics for a run.

        Args:
            run_id(str): The id of the run to get metrics for
            since (datetime): Only get metrics newer than this date
                Default: ``None``
            summarize (int): If set, metrics are summarized to at most this
            many entries by averaging the metrics. Default: ``None``

        Returns:
            A ``concurrent.futures.Future`` objects wrapping
            ``requests.response`` object. Get the result by calling
            ``return_value.result().json()``
        """
        return self._get_filtered_metrics(
            run_id=run_id,
            since=since,
            summarize=summarize,
            metric_filter=metric_filter,
            last_n=last_n,
        )

    def get_pod_metrics(self, pod_id, since=None, summarize=None):
        """ Get all metrics for a worker pod.

        Args:
            pod_id(str): The id of the pod to get metrics for
            since (datetime): Only get metrics newer than this date
                Default: ``None``
            summarize (int): If set, metrics are summarized to at most this
            many entries by averaging the metrics. Default: ``None``

        Returns:
            A ``concurrent.futures.Future`` objects wrapping
            ``requests.response`` object. Get the result by calling
            ``return_value.result().json()``
        """
        return self._get_filtered_metrics(
            pod_id=pod_id, since=since, summarize=summarize
        )

    def _get_filtered_metrics(
        self,
        pod_id=None,
        run_id=None,
        since=None,
        summarize=None,
        metric_filter=None,
        last_n=None,
        format=None,
    ):
        """Get metrics for a run or pod"""
        if pod_id is None and run_id is None:
            raise ValueError("Either pod_id or pod_id must be specified")

        pk = pod_id
        metric_type = "pod"

        if pk is None:
            pk = run_id
            metric_type = "run"

        params = {"metric_type": metric_type}

        if since is not None:
            params["since"] = str(since)

        if format is not None:
            params["format"] = format

        if summarize is not None:
            params["summarize"] = str(summarize)

        if metric_filter is not None:
            params["metric_filter"] = str(metric_filter)

        if last_n is not None:
            params["last_n"] = str(last_n)

        request_url = "{endpoint}metrics/{pk}".format(endpoint=self.endpoint, pk=pk)

        future = self.executor.submit(requests.get, request_url, params=params)

        return future

    def download_run_metrics(self, run_id, since=None, summarize=None):
        """ Get all metrics for a run as zip.

        Args:
            run_id(str): The id of the run to get metrics for
            since (datetime): Only get metrics newer than this date
                Default: ``None``
            summarize (int): If set, metrics are summarized to at most this
            many entries by averaging the metrics. Default: ``None``

        Returns:
            A ``concurrent.futures.Future`` objects wrapping
            ``requests.response`` object. Get the result by calling
            ``return_value.result().json()``
        """
        return self._get_filtered_metrics(
            run_id=run_id, since=since, summarize=summarize, format="zip"
        )

    def post_metric(
        self, run_id, name, value, cumulative=False, metadata="", date=None
    ):
        """ Save a metric to the master node for a run.

        Args:
            run_id(str): The id of the run to save a metric for
            name (str): The name of the metric, e.g. ``accuracy``
            value (Number): The metric value to save
            cumulative (bool, optional): Whether this metric is
                cumulative or not. Cumulative metrics are values
                that increment over time, i.e. ``current_calue =
                previous_value + value_difference``. Non-cumulative
                values or discrete values at a certain time. Default: ``False``
            metadata (dict): Optional metadata to attach to a metric.
                Default: ``None``
            date (datetime): The date the metric was gathered.
                Default: ``datetime.now``

        Returns:
            A ``concurrent.futures.Future`` objects wrapping
            ``requests.response`` object. Get the result by calling
            ``return_value.result().json()``
        """
        if date is None:
            date = datetime.datetime.utcnow()

        request_url = "{endpoint}metrics/".format(endpoint=self.endpoint)
        future = self.executor.submit(
            requests.post,
            request_url,
            data={
                "run_id": run_id,
                "name": name,
                "cumulative": cumulative,
                "date": str(date),
                "value": str(value),
                "metadata": metadata,
            },
            timeout=180,
        )
        return future

    def get_runs(self):
        """ Get all active, finished and failed benchmark runs

        Returns:
            A ``concurrent.futures.Future`` objects wrapping
            ``requests.response`` object. Get the result by calling
            ``return_value.result().json()``
        """
        request_url = "{endpoint}runs/".format(endpoint=self.endpoint)
        future = self.executor.submit(requests.get, request_url)

        return future

    def get_run(self, run_id):
        """ Get a specific benchmark run

        Args:
            run_id(str): The id of the run to get

        Returns:
            A ``concurrent.futures.Future`` objects wrapping
            ``requests.response`` object. Get the result by calling
            ``return_value.result().json()``
        """
        request_url = "{endpoint}runs/{run_id}".format(
            endpoint=self.endpoint, run_id=run_id
        )
        future = self.executor.submit(requests.get, request_url)

        return future

    def create_run(
        self,
        name,
        num_workers,
        num_cpus=2.0,
        max_bandwidth=1000,
        image=None,
        backend=None,
        custom_image_name=None,
        custom_image_command=None,
        custom_backend=None,
        run_all_nodes=False,
        gpu_enabled=False,
        light_target=False,
    ):
        """ Create a new benchmark run.

        Available official benchmarks can be found in
        the ``mlbench_core.api.MLBENCH_IMAGES`` dict.

        Args:
            name (str): The name of the run
            num_workers (int): The number of worker nodes to use
            num_cpus (float): The number of CPU Cores per worker to utilize.
                Default: ``2.0``
            max_bandwidth (int): Maximum bandwidth available for
                communication between worker nodes in mbps. Default: ``1000``
            image (str): Name of the official benchmark image to use (
                see ``mlbench_core.api.MLBENCH_IMAGES`` keys).
                Default: ``None``
            backend (str): Name of the backend to use (see ``mlbench_core.api.MLBENCH_BACKENDS``)
                Default: ``None``
            custom_image_name (str): The name of a custom Docker image
                to run. Can be a dockerhub or private Docker repository url.
                Default: ``None``
            custom_image_command (str): Command to run on the custom image.
                Default: ``None``
            custom_backend (str): Custom backend to use.
                Default: ``None``
            run_all_nodes (bool): Whether to run
                ``custom_image_command`` on all worker nodes or only the
                rank 0 node.
            gpu_enabled (bool): Enable GPU acceleration.
                Default: ``False``
            light_target (bool): Use light target goal
                Default: ``False``

        Returns:
            A ``concurrent.futures.Future`` objects wrapping
            ``requests.response`` object. Get the result by calling
            ``return_value.result().json()``
        """
        request_url = "{endpoint}runs/".format(endpoint=self.endpoint)
        data = {
            "name": name,
            "num_workers": num_workers,
            "num_cpus": num_cpus,
            "max_bandwidth": max_bandwidth,
            "gpu_enabled": "true" if gpu_enabled else "false",
            "light_target": "true" if light_target else "false",
            "run_all_nodes": "true" if run_all_nodes else "false",
        }

        if custom_image_name is not None:
            data["image_name"] = "custom_image"
            data["custom_image_name"] = custom_image_name
            data["custom_image_command"] = custom_image_command
        elif image in MLBENCH_IMAGES:
            data["image_name"] = MLBENCH_IMAGES[image][0]

            if MLBENCH_IMAGES[image][2]:
                data["gpu_enabled"] = "true" if gpu_enabled else "false"
        else:
            raise ValueError("Image {image} not found".format(image=image))

        assert not (
            (custom_backend is not None) and (backend is not None)
        ), "custom_backend and backend are mutually exclusive"
        if custom_backend is not None:
            data["backend"] = "custom_backend"
            data["custom_backend"] = custom_backend
        elif backend in MLBENCH_BACKENDS:
            data["backend"] = backend

        future = self.executor.submit(requests.post, request_url, data=data)
        return future

    def delete_run(self, run_id):
        """ Delete a benchmark run.

         Args:
            run_id(str): The id of the run to get

        Returns:
            A ``concurrent.futures.Future`` objects wrapping
            ``requests.response`` object. Get the result by calling
            ``return_value.result().json()``
        """
        request_url = "{endpoint}runs/{pk}".format(endpoint=self.endpoint, pk=run_id)

        future = self.executor.submit(requests.delete, request_url)
        return future

    def get_worker_pods(self):
        """ Get information on all worker nodes.

        Returns:
            A ``concurrent.futures.Future`` objects wrapping
            ``requests.response`` object. Get the result by calling
            ``return_value.result().json()``
        """
        request_url = "{endpoint}pods/".format(endpoint=self.endpoint)
        future = self.executor.submit(requests.get, request_url)

        return future

    def __enter__(self):
        self.executor.__enter__()
        return self

    def __exit__(self, *args):
        return self.executor.__exit__(*args)



import datetime
import os

from mlbench_core.api import ApiClient


class LogMetrics(object):
    """ Use to write metric values to the Dashboard API and to Trackers

    Caches API client for performance reasons
    """

    in_cluster = os.getenv("KUBERNETES_SERVICE_HOST") is not None

    if in_cluster:
        api = ApiClient()

    @staticmethod
    def log(run_id, rank, epoch, metric_name, value):
        """ Logs metrics to the Metrics API

        Currently only logs inside of a cluster

        Args:
            run_id (str): The id of the run in the dashboard
            rank (int): Rank of the current worker node
            epoch (float): The current epoch (fractional)
            metric_name (str): The name of the metric
            value (float / int / str): The metric value to write
            tracker(:obj:`mlbench_core.utils.Tracker`): The value Tracker
            time (float): The current time (used for Tracker)

        """

        if not LogMetrics.in_cluster:
            return

        metric_name = "{} @ {}".format(metric_name, rank)

        LogMetrics.api.post_metric(
            run_id,
            metric_name,
            value,
            metadata="{{rank: {}, epoch:{}}}".format(rank, epoch),
        )



import numpy as np
import torch
import torch.distributed as dist
from apex import amp
from mlbench_core.utils.pytorch.distributed import (
    AllReduceAggregation,
    AllReduceAggregationFP16,
)
from mlbench_core.utils.pytorch.distributed import DecentralizedAggregation
from torch.nn.utils import clip_grad_norm_
from torch.optim import Adam, SGD
from torch.optim.optimizer import Optimizer, required


class SparsifiedSGD(Optimizer):
    r"""Implements sparsified version of stochastic gradient descent.

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): learning rate
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        sparse_grad_size (int): Size of the sparsified gradients vector (
        default: 10).

    """

    def __init__(self, params, lr=required, weight_decay=0, sparse_grad_size=10):

        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = dict(lr=lr, weight_decay=weight_decay)

        super(SparsifiedSGD, self).__init__(params, defaults)

        self.__create_gradients_memory()
        self.__create_weighted_average_params()

        self.num_coordinates = sparse_grad_size

    def __create_weighted_average_params(self):
        r""" Create a memory to keep the weighted average of parameters in
        each iteration """
        for group in self.param_groups:
            for p in group["params"]:
                param_state = self.state[p]
                param_state["estimated_w"] = torch.zeros_like(p.data)
                p.data.normal_(0, 0.01)
                param_state["estimated_w"].copy_(p.data)

    def __create_gradients_memory(self):
        r""" Create a memory to keep gradients that are not used in each
        iteration """
        for group in self.param_groups:
            for p in group["params"]:
                param_state = self.state[p]
                param_state["memory"] = torch.zeros_like(p.data)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            weight_decay = group["weight_decay"]

            for p in group["params"]:

                if p.grad is None:
                    continue
                d_p = p.grad.data

                if weight_decay != 0:
                    d_p.add_(weight_decay, p.data)
                p.data.add_(-d_p)

        return loss

    def sparsify_gradients(self, param, lr):
        """ Calls one of the sparsification functions (random or blockwise)

        Args:
            random_sparse (bool): Indicates the way we want to make the
            gradients sparse
                (random or blockwise) (default: False)
            param (:obj: `torch.nn.Parameter`): Model parameter
        """
        if self.random_sparse:
            return self._random_sparsify(param, lr)
        else:
            return self._block_sparsify(param, lr)

    def _random_sparsify(self, param, lr):
        """ Sparsify the gradients vector by selecting 'k' of them randomly.

        Args:
            param (:obj: `torch.nn.Parameter`): Model parameter
            lr (float): Learning rate

        """

        self.state[param]["memory"] += param.grad.data * lr

        indices = np.random.choice(
            param.data.size()[1], self.num_coordinates, replace=False
        )
        sparse_tensor = torch.zeros(2, self.num_coordinates)

        for i, random_index in enumerate(indices):
            sparse_tensor[1, i] = self.state[param]["memory"][0, random_index]
            self.state[param]["memory"][0, random_index] = 0
        sparse_tensor[0, :] = torch.tensor(indices)

        return sparse_tensor

    def _block_sparsify(self, param, lr):
        """ Sparsify the gradients vector by selecting a block of them.

        Args:
            param (:obj: `torch.nn.Parameter`): Model parameter
            lr (float): Learning rate
        """

        self.state[param]["memory"] += param.grad.data * lr

        num_block = int(param.data.size()[1] / self.num_coordinates)

        current_block = np.random.randint(0, num_block)
        begin_index = current_block * self.num_coordinates

        end_index = begin_index + self.num_coordinates - 1
        output_size = 1 + end_index - begin_index + 1

        sparse_tensor = torch.zeros(1, output_size)
        sparse_tensor[0, 0] = begin_index
        sparse_tensor[0, 1:] = self.state[param]["memory"][
            0, begin_index : end_index + 1
        ]
        self.state[param]["memory"][0, begin_index : end_index + 1] = 0

        return sparse_tensor

    def update_estimated_weights(self, iteration, sparse_vector_size):
        """ Updates the estimated parameters

        Args:
            iteration (int): Current global iteration
            sparse_vector_size (int): Size of the sparse gradients vector
        """
        t = iteration
        for group in self.param_groups:
            for param in group["params"]:
                tau = param.data.size()[1] / sparse_vector_size
                rho = (
                    6
                    * ((t + tau) ** 2)
                    / ((1 + t) * (6 * (tau ** 2) + t + 6 * tau * t + 2 * (t ** 2)))
                )
                self.state[param]["estimated_w"] = (
                    self.state[param]["estimated_w"] * (1 - rho) + param.data * rho
                )

    def get_estimated_weights(self):
        """ Returns the weighted average parameter tensor """
        estimated_params = []
        for group in self.param_groups:
            for param in group["params"]:
                estimated_params.append(self.state[param]["estimated_w"])
        return estimated_params


class CentralizedSparsifiedSGD(SparsifiedSGD):
    r"""Implements centralized sparsified version of stochastic gradient
    descent.

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): Learning rate
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        sparse_grad_size (int): Size of the sparsified gradients vector (
        default: 10)
        random_sparse (bool): Whether select random sparsification (default:
        `False`)
        average_models (bool): Whether to average models together (default:
        `True`)

    """

    def __init__(
        self,
        params=None,
        lr=required,
        weight_decay=0,
        sparse_grad_size=10,
        random_sparse=False,
        average_models=True,
    ):
        if not params:
            raise ValueError('"params" not set for optimizer')
        self.average_models = average_models
        self.world_size = dist.get_world_size()
        self.random_sparse = random_sparse
        super(CentralizedSparsifiedSGD, self).__init__(
            params, lr, weight_decay, sparse_grad_size
        )

    def step(self, closure=None):
        """ Aggregates the gradients and performs a single optimization step.

            Arguments:
                closure (callable, optional): A closure that reevaluates the
                model and returns the loss.
        """

        loss = None

        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            weight_decay = group["weight_decay"]
            lr = group["lr"]

            for p in group["params"]:
                # Sparsify the gradients
                sparse_tensor = self.sparsify_gradients(p, lr)
                # Aggregate the gradients
                gathered_list = [
                    torch.zeros_like(sparse_tensor) for _ in range(self.world_size)
                ]
                dist.all_gather(gathered_list, sparse_tensor)
                p.grad.data = torch.zeros_like(p.grad.data)

                if self.random_sparse:
                    for grad_tensor in gathered_list:
                        for index in range(grad_tensor.size()[1]):
                            p.grad.data[0, int(grad_tensor[0, index])] += grad_tensor[
                                1, index
                            ]
                else:
                    for grad_tensor in gathered_list:
                        tensor_size = grad_tensor.size()[1]
                        begin = int(grad_tensor[0, 0])
                        p.grad.data[
                            0, begin : (begin + tensor_size - 1)
                        ] += grad_tensor[0, 1:]

                if self.average_models:
                    p.grad.data /= self.world_size

                if p.grad is None:
                    continue
                d_p = p.grad.data

                if weight_decay != 0:
                    d_p.add_(weight_decay, p.data)
                p.data.add_(-d_p)

        return loss


class DecentralizedSGD(SGD):
    r"""Implements decentralized stochastic gradient descent (optionally
    with momentum).

    Args:
        rank (int): rank of current process in the network
        neighbors (list): list of ranks of the neighbors of current process
        model (:obj:`nn.Module`): model which contains parameters for SGD
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)
        average_models (bool): Whether to average models together. (default: `True`)
        use_cuda (bool): Whether to use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer instead of all layers at once
    """

    def __init__(
        self,
        rank=None,
        neighbors=None,
        model=None,
        lr=required,
        momentum=0,
        dampening=0,
        weight_decay=0,
        nesterov=False,
        average_models=True,
        use_cuda=False,
        by_layer=False,
    ):
        if not rank:
            raise ValueError('"rank" not set for optimizer')
        if not neighbors:
            raise ValueError('"neighbors" not set for optimizer')
        if not model:
            raise ValueError('"model" not set for optimizer')
        super(DecentralizedSGD, self).__init__(
            model.parameters(), lr, momentum, dampening, weight_decay, nesterov
        )

        if average_models:
            self.agg_mode = "avg"
        else:
            raise NotImplementedError("Only average model is supported right now.")

        self.model = model
        self.agg = DecentralizedAggregation(
            rank, neighbors, use_cuda=use_cuda
        ).agg_model(by_layer=by_layer)

    def step(self, closure=None):
        """ Aggregates the gradients and performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = super(DecentralizedSGD, self).step(closure=closure)
        # Averaging the model after updating the gradient separately.
        self.agg(self.model, self.agg_mode)
        return loss


class CentralizedSGD(SGD):
    r"""Implements centralized stochastic gradient descent (optionally with
    momentum).

    Args:
        world_size (int): Size of the network
        model (:obj:`nn.Module`): Model which contains parameters for SGD
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)
        average_models (bool): Whether to average models together. (default: `True`)
        use_cuda (bool): Whether to use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer instead of all layers at once
    """

    def __init__(
        self,
        world_size=None,
        model=None,
        lr=required,
        momentum=0,
        dampening=0,
        weight_decay=0,
        nesterov=False,
        average_models=True,
        use_cuda=False,
        by_layer=False,
    ):
        if not world_size:
            raise ValueError('"world_size" not set for optimizer')
        if not model:
            raise ValueError('"model" not set for optimizer')
        super(CentralizedSGD, self).__init__(
            model.parameters(), lr, momentum, dampening, weight_decay, nesterov
        )
        if average_models:
            self.agg_mode = "avg"
        else:
            raise NotImplementedError("Only average model is supported right now.")

        self.model = model
        self.agg = AllReduceAggregation(
            world_size=world_size, use_cuda=use_cuda
        ).agg_grad(by_layer=by_layer)

    def step(self, closure=None):
        """ Aggregates the gradients and performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        self.agg(self.model, self.agg_mode)
        loss = super(CentralizedSGD, self).step(closure=closure)
        return loss


class SignSGD(SGD):
    r"""Implements sign stochastic gradient descent (optionally with momentum).

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)
        average_models (bool): Whether to average models together. (default:
        `True`)

    """

    def step(self, closure=None):
        """ Aggregates the gradients and performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group["weight_decay"]
            momentum = group["momentum"]
            dampening = group["dampening"]
            nesterov = group["nesterov"]

            for p in group["params"]:
                if p.grad is None:
                    continue
                d_p = p.grad.data
                if weight_decay != 0:
                    d_p.add_(weight_decay, p.data)
                if momentum != 0:
                    param_state = self.state[p]
                    if "momentum_buffer" not in param_state:
                        buf = param_state["momentum_buffer"] = torch.zeros_like(p.data)
                        buf.mul_(momentum).add_(d_p)
                    else:
                        buf = param_state["momentum_buffer"]
                        buf.mul_(momentum).add_(1 - dampening, d_p)
                    if nesterov:
                        d_p = d_p.add(momentum, buf)
                    else:
                        d_p = buf

                # Update with the sign
                p.data.add_(-group["lr"], torch.sign(d_p))

        return loss


class CentralizedAdam(Adam):
    r"""Implements centralized Adam algorithm.

    Args:
        world_size (int): Size of the network
        model (:obj:`nn.Module`): Model which contains parameters for Adam
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper `On the Convergence of Adam and Beyond`_
            (default: False)
        average_models (bool): Whether to average models together. (default: `True`)
        use_cuda (bool): Whether to use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer instead of all layers at once
    """

    def __init__(
        self,
        world_size=None,
        model=None,
        lr=1e-3,
        betas=(0.9, 0.999),
        eps=1e-8,
        weight_decay=0,
        amsgrad=False,
        average_models=True,
        use_cuda=False,
        by_layer=False,
    ):
        if not world_size:
            raise ValueError('"world_size" not set for optimizer')
        if not model:
            raise ValueError('"model" not set for optimizer')
        super(CentralizedAdam, self).__init__(
            model.parameters(), lr, betas, eps, weight_decay, amsgrad
        )
        if average_models:
            self.agg_mode = "avg"
        else:
            raise NotImplementedError("Only average model is supported right now.")

        self.model = model
        self.agg = AllReduceAggregation(
            world_size=world_size, use_cuda=use_cuda
        ).agg_grad(by_layer=by_layer)

    def step(self, closure=None):
        """ Aggregates the gradients and performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        self.agg(self.model, self.agg_mode)
        loss = super(CentralizedAdam, self).step(closure=closure)
        return loss


class FP32Optimizer:
    """
    Standard optimizer, computes backward and applies weight update.
    """

    def __init__(self, model, optimizer, grad_clip=None):
        """
        Constructor for the Fp32Optimizer

        Args:
            model (torch.nn.Module): Model
            optimizer: The underlying optimizer
            grad_clip (float): Coefficient for gradient clipping (max L2 norm of gradients)
        """
        self.model = model
        self.params = model.parameters()
        self.grad_clip = grad_clip
        self.optimizer = optimizer

    def step(self, closure=None):
        """
        Performs one step of the optimizer.
        """
        if self.grad_clip != float("inf"):
            clip_grad_norm_(self.params, self.grad_clip)
        return self.optimizer.step(closure=closure)

    def backward_loss(self, loss):
        loss.backward()

    def zero_grad(self):
        self.optimizer.zero_grad()


class AMPOptimizer:
    """
    Optimizer compatible with AMP.
    Uses AMP to apply loss scaling, computes backward and applies weight
    update.
    """

    def __init__(
        self,
        model,
        optimizer,
        grad_clip=None,
        loss_scale=8192,
        dls_upscale_interval=128,
        average_models=True,
        world_size=1,
        use_cuda=False,
        by_layer=False,
        use_horovod=False,
    ):
        """
        Constructor for the AMPOptimizer

        Args:
            model (torch.nn.Module): Model
            optimizer: The underlying optimizer
            grad_clip (float): Coefficient for gradient clipping, max L2 norm of the gradients
            loss_scale:
            dls_upscale_interval:
        """
        self.model = model
        self.grad_clip = grad_clip
        self.optimizer = optimizer
        loss_scaler = amp._amp_state.loss_scalers[0]
        loss_scaler._loss_scale = loss_scale
        loss_scaler._scale_seq_len = dls_upscale_interval

        if average_models:
            self.agg_mode = "avg"
        else:
            raise NotImplementedError("Only average model is supported right now.")

        if use_horovod:
            self.agg = AllReduceAggregationFP16(
                world_size=world_size, use_cuda=use_cuda
            ).agg_grad(by_layer=by_layer)
        else:
            self.agg = AllReduceAggregation(
                world_size=world_size, use_cuda=use_cuda
            ).agg_grad(by_layer=by_layer)

    def backward_loss(self, loss):
        with amp.scale_loss(loss, self.optimizer) as scaled_loss:
            scaled_loss.backward()

    def step(self, closure=None):
        """
        Performs one step of the optimizer.
        """
        if self.grad_clip != float("inf"):
            clip_grad_norm_(amp.master_params(self.optimizer), self.grad_clip)

        self.agg(self.model, self.agg_mode)
        self.optimizer.step(closure=closure)

    def zero_grad(self):
        self.optimizer.zero_grad()


optimizers = {
    "centralized_sparsified_sgd": CentralizedSparsifiedSGD,
    "decentralized_sgd": DecentralizedSGD,
    "centralized_sgd": CentralizedSGD,
    "sign_sgd": SignSGD,
    "centralized_adam": CentralizedAdam,
}


def get_optimizer(optimizer, **kwargs):
    r"""Returns an object of the class specified with the argument `optimizer`.

        Args:
            optimizer (str): name of the optimizer
            **kwargs (dict, optional): additional optimizer-specific parameters. For the list of supported parameters
                for each optimizer, please look at its documentation.
        """
    return optimizers[optimizer](**kwargs)



from .controlflow import TrainValidation, train_round, validation_round
from .checkpoints_evaluation import CheckpointsEvaluationControlFlow

__all__ = [
    "TrainValidation",
    "CheckpointsEvaluationControlFlow",
    "train_round",
    "validation_round",
]



import torch
import torch.nn as nn


class RNNLM(nn.Module):
    """Container module with an encoder, a recurrent module, and a decoder."""

    def __init__(
        self,
        ntoken,
        ninp,
        nhid,
        nlayers,
        dropout=0.5,
        tie_weights=False,
        weight_norm=False,
    ):
        super(RNNLM, self).__init__()

        # define conf.
        self.nhid = nhid
        self.nlayers = nlayers

        # define layers.
        self.drop = nn.Dropout(dropout)
        self.encoder = nn.Embedding(ntoken, ninp)
        self.rnn = nn.LSTM(ninp, nhid, nlayers, dropout=dropout, batch_first=False)
        self.decoder = nn.Linear(nhid, ntoken)

        # Optionally tie weights as in:
        # "Using the Output Embedding to Improve Language Models" (Press & Wolf 2016)
        # https://arxiv.org/abs/1608.05859
        # and
        # "Tying Word Vectors and Word Classifiers: A Loss Framework for Language Modeling" (Inan et al. 2016)
        # https://arxiv.org/abs/1611.01462
        if tie_weights:
            if nhid != ninp:
                raise ValueError(
                    "When using the tied flag, nhid must be equal to emsize"
                )
            self.decoder.weight = self.encoder.weight

        self.init_weights()

    def init_weights(self):
        initrange = 0.1
        self.encoder.weight.data.uniform_(-initrange, initrange)
        self.decoder.bias.data.zero_()
        self.decoder.weight.data.uniform_(-initrange, initrange)

    def forward(self, input, hidden):
        emb = self.drop(self.encoder(input))
        output, hidden = self.rnn(emb, hidden)
        output = self.drop(output)
        decoded = self.decoder(
            output.view(output.size(0) * output.size(1), output.size(2))
        )
        return decoded.view(output.size(0), output.size(1), decoded.size(1)), hidden

    def init_hidden(self, bsz):
        weight = next(self.parameters())
        return (
            weight.new_zeros(self.nlayers, bsz, self.nhid),
            weight.new_zeros(self.nlayers, bsz, self.nhid),
        )

    def repackage_hidden(self, h):
        """Wraps hidden states in new Tensors, to detach them from their history."""
        if isinstance(h, torch.Tensor):
            return h.detach()
        else:
            return tuple(self.repackage_hidden(v) for v in h)



import os
from collections import defaultdict
from functools import partial

import sacremoses
import subword_nmt.apply_bpe
import torch
from mlbench_core.dataset.translation.pytorch import config


def _pad_vocabulary(vocab, math):
    """
    Pads vocabulary to a multiple of 'pad' tokens.

    Args:
        vocab (list): list with vocabulary
        math (str): Math precision. either `fp_16`, `manual_fp16` or `fp32`

    Returns:
        list: padded vocabulary
    """
    if math == "fp16" or math == "manual_fp16":
        pad = 8
    elif math == "fp32":
        pad = 1
    else:
        raise NotImplementedError()

    vocab_size = len(vocab)
    padded_vocab_size = (vocab_size + pad - 1) // pad * pad
    for i in range(0, padded_vocab_size - vocab_size):
        token = f"madeupword{i:04d}"
        vocab.append(token)
    assert len(vocab) % pad == 0
    return vocab


def _parse_vocab(vocab_fname, math_precision):
    """
    Parses the vocabulary file
    Args:
        vocab_fname (str): Vocab file name
        math_precision (str): Math mode. either `fp_16`, `manual_fp16` or `fp32`

    Returns:
        list, dict, dict: List of words, id-to-token and token-to-id dictionaries
    """
    vocab = [config.PAD_TOKEN, config.UNK_TOKEN, config.BOS_TOKEN, config.EOS_TOKEN]
    with open(vocab_fname, encoding="utf-8") as vfile:
        for line in vfile:
            vocab.append(line.strip())

    vocab = _pad_vocabulary(vocab, math_precision)

    tok2idx = defaultdict(partial(int, config.UNK))
    for idx, token in enumerate(vocab):
        tok2idx[token] = idx

    idx2tok = {}
    for key, value in tok2idx.items():
        idx2tok[value] = key

    return vocab, idx2tok, tok2idx


class WMT14Tokenizer:
    """Tokenizer Class for WMT14 that uses the whole vocabulary

    Args:
        base_dir (str): Base directory for files
        lang (dict): With keys `src` and `trg` designating source and target language
        math_precision (str): Math precision
        separator:
    """

    def __init__(
        self, base_dir, lang=None, math_precision=None, separator="@@",
    ):
        self.separator = separator
        self.lang = lang

        # base_dir = os.path.join(base_dir, "wmt14")
        bpe_fname = os.path.join(base_dir, config.BPE_CODES)
        vocab_fname = os.path.join(base_dir, config.VOCAB_FNAME)

        if bpe_fname:
            with open(bpe_fname, "r", encoding="utf-8") as bpe_codes:
                self.bpe = subword_nmt.apply_bpe.BPE(bpe_codes)

        if vocab_fname:
            tmp = _parse_vocab(vocab_fname, math_precision)

            vocab, self.idx2tok, self.tok2idx = tmp
            self.vocab_size = len(vocab)

        if lang:
            self.moses_tokenizer = sacremoses.MosesTokenizer(lang["src"])
            self.moses_detokenizer = sacremoses.MosesDetokenizer(lang["trg"])

    def parse_line(self, line):
        return torch.Tensor(self.segment(self.preprocess(line)))

    def segment(self, line):
        """
        Tokenizes single sentence and adds special BOS and EOS tokens.

        Args:
            line (str): sentence

        Returns:
            list representing tokenized sentence
        """
        entry = [self.tok2idx[i] for i in line]
        entry = [config.BOS] + entry + [config.EOS]
        return entry

    def detokenize_bpe(self, inp, delim=" "):
        """
        Detokenizes single sentence and removes token separator characters.

        Args:
            inp (list): sequence of tokens
            delim (str): tokenization delimiter

        Returns:
            string representing detokenized sentence
        """
        detok = delim.join([self.idx2tok[idx] for idx in inp])
        detok = detok.replace(self.separator + " ", "")
        detok = detok.replace(self.separator, "")

        detok = detok.replace(config.BOS_TOKEN, "")
        detok = detok.replace(config.EOS_TOKEN, "")
        detok = detok.replace(config.PAD_TOKEN, "")
        detok = detok.strip()
        return detok

    def detokenize_moses(self, inp):
        output = self.moses_detokenizer.detokenize(inp.split())
        return output

    def detokenize(self, inp):
        detok_bpe = self.detokenize_bpe(inp)
        output = self.detokenize_moses(detok_bpe)
        return output

    def preprocess(self, x):
        return x.strip().split()



"""VGG11/13/16/19 in Pytorch.

From https://github.com/kuangliu/pytorch-cifar."""
import torch
import torch.nn as nn


cfg = {
    "VGG11": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "VGG13": [64, 64, "M", 128, 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "VGG16": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        "M",
    ],
    "VGG19": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        512,
        "M",
    ],
}


class VGG(nn.Module):
    def __init__(self, vgg_name):
        super(VGG, self).__init__()
        self.features = self._make_layers(cfg[vgg_name])
        self.classifier = nn.Linear(512, 10)

    def forward(self, x):
        out = self.features(x)
        out = out.view(out.size(0), -1)
        out = self.classifier(out)
        return out

    def _make_layers(self, cfg):
        layers = []
        in_channels = 3
        for x in cfg:
            if x == "M":
                layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
            else:
                layers += [
                    nn.Conv2d(in_channels, x, kernel_size=3, padding=1),
                    nn.BatchNorm2d(x),
                    nn.ReLU(inplace=True),
                ]
                in_channels = x
        layers += [nn.AvgPool2d(kernel_size=1, stride=1)]
        return nn.Sequential(*layers)


def test():
    net = VGG("VGG11")
    x = torch.randn(2, 3, 32, 32)
    y = net(x)
    print(y.size())



import time
from collections import defaultdict

from .log_metrics import LogMetrics


class AverageMeter(object):
    """Computes and stores the average and current value."""

    def __init__(self):
        self.reset()

    def reset(self):
        """Reset all stats."""
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        """Update stats given input val and n."""
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


class Tracker(object):
    """A class to track running stats and metrics.

    Also responsible for posting metrics to the API/Dashboard

    Args:
        metrics (list): List of metrics objects
        run_id (int): The id of the current run
        rank (int): The rank of this worker node
        goal(func): A task goal to check for when logging metrics"""

    def __init__(
        self,
        metrics,
        run_id,
        rank,
        goal=None,
        communication_steps=["opt_step"],
        compute_steps=["fwd_pass", "comp_loss", "backprop"],
    ):
        self.batch_times = []
        self.validation_times = []
        self.epoch_stats = {}
        self.epoch_metrics = defaultdict(list)
        self.records = []
        self.history = []
        self.cumulative_train_time = []
        self.cumulative_compute_time = []
        self.cumulative_communication_time = []
        self.cumulative_val_time = []
        self.best_epoch = 0
        self.current_epoch = 0
        self.best_metric_value = 0
        self.is_training = True

        self.communication_steps = []
        self.compute_steps = []

        self.train_prefix = "train_"
        self.val_prefix = "val_"

        self.start_time = None
        self.metrics = metrics
        self.run_id = run_id
        self.rank = rank
        self.reset_epoch_stats()
        self.goal = goal
        self.goal_reached = False

        self.primary_metric = metrics[0]

        if not isinstance(communication_steps, list):
            communication_steps = [communication_steps]

        self.communication_steps = communication_steps

        if not isinstance(compute_steps, list):
            compute_steps = [compute_steps]

        self.compute_steps = compute_steps

    def start(self):
        """ Starts Tracking """
        if self.start_time is not None:
            raise Exception("Tracking already started")

        self.start_time = time.time()

    def train(self):
        """ Switch Tracker to training mode"""
        self.is_training = True

    def validation(self):
        """Switch Tracker to validation mode"""
        self.is_training = False

    def validation_start(self):
        """Start validation step & timer"""
        self.validation_times = [("start", time.time())]

    def validation_end(self):
        """End Validation step"""
        self.validation_times = [("end", time.time())]
        self.cumulative_val_time.append(
            self.validation_times[-1][1] - self.validation_times[0][1]
        )

    def batch_start(self):
        """Start a training batch"""
        self.batch_times = []
        self.record_batch_step("start")

    def batch_end(self):
        """End a training batch and calculate time spent"""
        self.record_batch_step("end")

        if len(self.batch_times) > 1:
            metric = "CumulativeTrainTimeEpoch"

            self.cumulative_train_time.append(
                self.batch_times[-1][1] - self.batch_times[0][1]
            )
            time_diff = self.cumulative_train_time[-1]

            # batch times
            self.batch_times.sort(key=lambda x: x[1])

            tracker_stats = list(zip(self.batch_times[:-1], self.batch_times[1:]))

            if self.compute_steps:
                compute_times = [
                    t[1][1] - t[0][1]
                    for t in tracker_stats
                    if t[1][0] in self.compute_steps
                ]
                self.cumulative_compute_time.append(sum(compute_times))

            if self.communication_steps:
                communication_times = [
                    t[1][1] - t[0][1]
                    for t in tracker_stats
                    if t[1][0] in self.communication_steps
                ]
                self.cumulative_communication_time.append(sum(communication_times))

            if len(self.epoch_metrics[metric]) < self.current_epoch + 1:
                self.epoch_metrics[metric].append(0.0)

            self.epoch_metrics[metric][self.current_epoch] += time_diff

            for (_, t1), (name, t2) in tracker_stats:
                time_diff = t2 - t1

                if len(self.epoch_metrics[name]) < self.current_epoch + 1:
                    self.epoch_metrics[name].append(0.0)

                self.epoch_metrics[name][self.current_epoch] += time_diff

    def epoch_end(self):
        """Ends a training epoch and logs epoch metrics"""
        for k, v in dict(self.epoch_metrics).items():
            LogMetrics.log(self.run_id, self.rank, self.current_epoch, k, v[-1])

        self.current_epoch += 1

        self.reset_epoch_stats()

    def reset_epoch_stats(self):
        """ Resets all epoch stats"""
        stat_names = [k for k in ["loss"] + [m.name for m in self.metrics]]

        stat_names = [self.train_prefix + k for k in stat_names] + [
            self.val_prefix + k for k in stat_names
        ]

        self.epoch_stats = {k: AverageMeter() for k in stat_names}

    def record_batch_step(self, name):
        """Records a specific batch step for timing (e.g. "backpropagation" or "model_forward

        Args:
            name (str): The name of the step to record"""
        self.batch_times.append((name, time.time()))

    def record_stat(self, name, value, n=1, log_to_api=False):
        """Records a stat value

        Args:
            name (str): The name of the stat to record
            value (number): The stat value to record
            n (int): The number of individual samples this stat is made up of
                     in case of an average value
            log_to_api (bool): Whether to submit the stat to the Dashboard API, default:False
        """
        prefix = self.train_prefix

        if not self.is_training:
            prefix = self.val_prefix

        name = prefix + name

        if name not in self.epoch_stats:
            self.epoch_stats[name] = AverageMeter()

        self.epoch_stats[name].update(value, n)

        self.history.append((self.run_id, self.rank, name, value, time.time()))

        if log_to_api:
            LogMetrics.log(self.run_id, self.rank, self.current_epoch, name, value)

        if self.goal and self.rank == 0:
            goal_result = self.goal(name, value, self)

            if goal_result is not None and not self.goal_reached:
                self.goal_reached = True
                print("goal reached1")
                print(log_to_api)
                print(goal_result)

                if log_to_api:
                    LogMetrics.log(
                        self.run_id,
                        self.rank,
                        self.current_epoch,
                        "TaskResult",
                        goal_result,
                    )

                    LogMetrics.log(
                        self.run_id,
                        self.rank,
                        self.current_epoch,
                        "TotalCumulativeTrainTime",
                        self.get_total_train_time(),
                    )

                    metrics = dict(self.epoch_metrics).items()
                    metrics = sorted(metrics, key=lambda k: k[0])

                    for k, v in metrics:
                        LogMetrics.log(
                            self.run_id,
                            self.rank,
                            self.current_epoch,
                            "global_cum_{}".format(k),
                            sum(v),
                        )

    def record_loss(self, value, n=1, log_to_api=False):
        """Records a loss value

        Args:
            value (number): The stat value to record
            n (int): The number of individual losses this stat is made up of
                     in case of an average value (e.g. if batch loss use batch size)
            log_to_api (bool): Whether to submit the stat to the Dashboard API, default:False
        """
        self.record_stat("loss", value, n, log_to_api)

    def record_metric(self, metric, value, n=1, log_to_api=False):
        """Records a metric value

        Args:
            metric (`obj`:metric): The metric to log
            value (number): The stat value to record
            n (int): The number of individual samples this stat is made up of
                     in case of an average value
            log_to_api (bool): Whether to submit the stat to the Dashboard API, default:False
        """
        self.record_stat(metric.name, value, n, log_to_api)

        if metric.name == self.primary_metric.name and not self.is_training:
            self.update_primary_metric(value)

    def update_primary_metric(self, new_metric_value):
        """ Updates the primary (main) metric

        Args:
            new_metric_value (number): The new value of the metric
        """

        if new_metric_value > self.best_metric_value:
            self.best_metric_value = new_metric_value
            self.best_epoch = self.current_epoch

    def is_best(self):
        """ Whether the current epoch is the best epoch so far"""
        return self.current_epoch == self.best_epoch

    def get_total_train_time(self):
        return sum(self.cumulative_train_time)

    def get_total_compute_time(self):
        return sum(self.cumulative_compute_time)

    def get_total_communication_time(self):
        return sum(self.cumulative_communication_time)

    def get_total_val_time(self):
        return sum(self.cumulative_val_time)

    def __str__(self):
        """ String representation of current epoch"""
        prefix = self.train_prefix

        if not self.is_training:
            prefix = self.val_prefix

        # loss
        str_builder = ["loss={:6.2e}".format(self.epoch_stats[prefix + "loss"].avg)]

        # metrics
        for metric in self.metrics:
            str_builder.append(
                "{} {:.2e}".format(
                    metric.name, self.epoch_stats[prefix + metric.name].avg
                )
            )

        # batch times
        self.batch_times.sort(key=lambda x: x[1])

        tracker_stats = zip(self.batch_times[:-1], self.batch_times[1:])

        str_builder.append(
            " | ".join(
                "{} {:.1e}".format(name, t2 - t1)
                for (_, t1), (name, t2) in tracker_stats
            )
        )

        return " | ".join(str_builder)
